# import importlib
# # https://snarky.ca/lazy-importing-in-python-3-7/


# def lazy_import(importer_name: str, to_import: str):
#     """Return the importing module and a callable for lazy importing.

#     The module named by importer_name represents the module performing the
#     import to help facilitate resolving relative imports.

#     to_import is an iterable of the modules to be potentially imported (absolute
#     or relative). The `as` form of importing is also supported,
#     e.g. `pkg.mod as spam`.

#     This function returns a tuple of two items. The first is the importer
#     module for easy reference within itself. The second item is a callable to be
#     set to `__getattr__`.
#     """
#     module = importlib.import_module(importer_name)
#     import_mapping = {}
#     for name in to_import:
#         importing, _, binding = name.partition(' as ')
#         if not binding:
#             _, _, binding = importing.rpartition('.')
#         import_mapping[binding] = importing

#     def __getattr__(name):
#         if name not in import_mapping:
#             message = f'module {importer_name!r} has no attribute {name!r}'
#             raise AttributeError(message)
#         importing = import_mapping[name]
#         # imortlib.import_module() implicitly sets submodules on this module as
#         # appropriate for direct imports.
#         imported = importlib.import_module(importing,
#                                            module.__spec__.parent)
#         setattr(module, name, imported)
#         return imported

#     return module, __getattr__
# try:
#     mod, __getattr__ = lazy_import(__name__, {
#         '.lo as mLo',
#         '.info as mInfo',
#         '.xml_util as mXML',
#         '.file_io as mFileIO',
#         '.props as mProps',
#         'gui as mGUI',
#         ".images_lo as mImgLo"
#         })
# except AttributeError:
#     pass


# coding: utf-8
# Python conversion of Info.java by Andrew Davison, ad@fivedots.coe.psu.ac.th
# See Also: https://fivedots.coe.psu.ac.th/~ad/jlop/
from __future__ import annotations
import os
from typing import cast, overload
from enum import IntEnum


from ..events.event_singleton import _Events
from ..events.lo_named_event import LoNamedEvent
from ..utils import lo as mLo
from ..utils import info as mInfo

from ..meta.static_meta import StaticProperty, classproperty
from ..utils.type_var import DocOrText, DocOrCursor
from ..exceptions import ex as mEx
from ..events.event_singleton import _Events
from ..events.args.event_args import EventArgs
from ..events.args.cancel_event_args import CancelEventArgs
from ..events.write_named_event import WriteNamedEvent


# if not _DOCS_BUILDING and not _ON_RTD:
from com.sun.star.beans import XPropertySet
from com.sun.star.container import XIndexAccess
from com.sun.star.frame import XModel
from com.sun.star.i18n import XBreakIterator
from com.sun.star.text import XParagraphCursor
from com.sun.star.text import XSentenceCursor
from com.sun.star.text import XText
from com.sun.star.text import XTextCursor
from com.sun.star.text import XTextDocument
from com.sun.star.text import XTextRange
from com.sun.star.text import XTextRangeCompare
from com.sun.star.text import XTextViewCursor
from com.sun.star.text import XTextViewCursorSupplier
from com.sun.star.text import XWordCursor
from com.sun.star.view import XSelectionSupplier

from ooo.dyn.i18n.word_type import WordTypeEnum as WordTypeEnum
from ooo.dyn.i18n.boundary import Boundary  # struct
from ooo.dyn.lang.locale import Locale  # struct


class Selection(metaclass=StaticProperty):
    """Selection Framework"""

    class CompareEnum(IntEnum):
        """Compare Enumeration"""

        AFTER = 1
        BEFORE = -1
        EQUAL = 0

    @staticmethod
    def is_anything_selected(text_doc: XTextDocument) -> bool:
        """
        Determine if anything is selected.

        If Write document is not visible this method returns false.

        Args:
            text_doc (XTextDocument): Text Document

        Returns:
            bool: True if anything in the document is selected: Otherwise, False

        Note:
            Writer must be visible for this method or ``False`` is always returned.
        """

        model = mLo.Lo.qi(XModel, text_doc, True)

        o_selections = model.getCurrentSelection()
        if o_selections is None:
            return False
        index_access = mLo.Lo.qi(XIndexAccess, o_selections)
        if index_access is None:
            return False
        count = int(index_access.getCount())
        if count == 0:
            return False
        elif count > 1:
            return True
        else:
            # There is only one selection so obtain the first selection
            o_sel = mLo.Lo.qi(XTextRange, index_access.getByIndex(0))
            if o_sel is None:
                return False
            o_text = o_sel.getText()
            # Create a text cursor that covers the range and then see if it is collapsed
            o_cursor = o_text.createTextCursorByRange(o_sel)
            if not o_cursor.isCollapsed():
                return True

        return False

    @staticmethod
    def get_selected_text_range(text_doc: XTextDocument) -> XTextRange | None:
        """
        Gets the text range for current selection

        Args:
            text_doc (XTextDocument): Text Document

        Raises:
            MissingInterfaceError: If unable to obtain required interface.

        Returns:
            XTextRange | None: If no selection is made then None is returned; Otherwise, Text Range.

        Note:
            Writer must be visible for this method or ``None`` is returned.
        """
        model = mLo.Lo.qi(XModel, text_doc, True)

        o_selections: XIndexAccess = model.getCurrentSelection()
        if not o_selections:
            return None
        count = int(o_selections.getCount())
        if count == 0:
            return None
        return mLo.Lo.qi(XTextRange, o_selections.getByIndex(0), True)

    @classmethod
    def get_selected_text_str(cls, text_doc: XTextDocument) -> str:
        """
        Gets the first selection text for Document

        Args:
            text_doc (XTextDocument): Text Document

        Returns:
            str: Selected text or empty string.

        Note:
            Writer must be visible for this method or empty string is returned.
        """
        rng = cls.get_selected_text_range(text_doc=text_doc)
        if rng is None:
            return ""
        return rng.getString()

    @classmethod
    def compare_cursor_ends(cls, c1: XTextRange, c2: XTextRange) -> Selection.CompareEnum:
        """
        Compares two cursors ranges end positions

        Args:
            c1 (XTextRange): first cursor range
            c2 (XTextRange): second cursor range

        Raises:
            Exception: if comparison fails

        Returns:
            CompareEnum: Compare result.
            :py:attr:`.CompareEnum.BEFORE` if ``c1`` end position is before ``c2`` end position.
            :py:attr:`.CompareEnum.EQUAL` if ``c1`` end position is equal to ``c2`` end position.
            :py:attr:`.CompareEnum.AFTER` if ``c1`` end position is after ``c2`` end position.
        """
        range_compare = cast(XTextRangeCompare, cls.text_range_compare)
        i = range_compare.compareRegionEnds(c1, c2)
        if i == 1:
            return cls.CompareEnum.BEFORE
        if i == -1:
            return cls.CompareEnum.AFTER
        if i == 0:
            return cls.CompareEnum.EQUAL
        # if no valid result raise error
        msg = "get_cursor_compare_ends() unable to get a valid compare result"
        raise Exception(msg)

    @classmethod
    def range_len(cls, text_doc: XTextDocument, o_sel: XTextRange) -> int:
        """
        Gets the distance between range start and range end.

        Args:
            o_sel (XTextRange): first cursor range
            o_text (object): XText object, usually document text object

        Returns:
            int: length of range

        Note:
            All characters are counted including paragraph breaks.
            In Writer it will display selected characters however,
            paragraph breaks are not counted.
        """
        i = 0
        if o_sel.isCollapsed():
            return i
        o_text = mLo.Lo.qi(XText, text_doc, True)
        l_cursor = cls.get_left_cursor(o_sel=o_sel, o_text=o_text)
        r_cursor = cls.get_right_cursor(o_sel=o_sel, o_text=o_text)
        if cls.compare_cursor_ends(c1=l_cursor, c2=r_cursor) < cls.CompareEnum.EQUAL:

            while cls.compare_cursor_ends(c1=l_cursor, c2=r_cursor) != cls.CompareEnum.EQUAL:
                l_cursor.goRight(1, False)
                i += 1
        return i

    # region ------------- model cursor methods ------------------------

    @classmethod
    def get_text_cursor_props(cls, text_doc: XTextDocument) -> XPropertySet:
        """
        Gets properties for document cursor

        Args:
            text_doc (XTextDocument): Text Document

        Raises:
            MissingInterfaceError: If unable to obtain XPropertySet interface from cursor.

        Returns:
            XPropertySet: Properties
        """
        cursor = cls.get_cursor(text_doc)
        props = mLo.Lo.qi(XPropertySet, cursor, True)
        return props

    # region    get_cursor()
    @overload
    @classmethod
    def get_cursor(cls, cursor_obj: DocOrCursor) -> XTextCursor:
        """
        Gets text cursor

        Args:
            cursor_obj (DocOrCursor): Text Document or Text Cursor

        Returns:
            XTextCursor: Cursor
        """
        ...

    @overload
    @classmethod
    def get_cursor(cls, rng: XTextRange, txt: XText) -> XTextCursor:
        """
        Gets text cursor

        Args:
            rng (XTextRange): Text Range Instance
            txt (XText): Text Instance

        Returns:
            XTextCursor: Cursor
        """
        ...

    @overload
    @classmethod
    def get_cursor(cls, rng: XTextRange, text_doc: XTextDocument) -> XTextCursor:
        """
        Gets text cursor

        Args:
            rng (XTextRange): Text Range instance
            text_doc (XTextDocument): Text Document instance

        Returns:
            XTextCursor: Cursor
        """
        ...

    @classmethod
    def get_cursor(cls, *args, **kwargs) -> XTextCursor:
        """
        Gets text cursor

        Args:
            cursor_obj (DocOrCursor): Text Document or Text View Cursor
            rng (XTextRange): Text Range Instance
            text_doc (XTextDocument): Text Document instance

        Raises:
            CursorError: If Unable to get cursor

        Returns:
            XTextCursor: Cursor
        """
        ordered_keys = (1, 2)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("cursor_obj", "rng", "txt", "text_doc")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("get_cursor() got an unexpected keyword argument")

            keys = ("cursor_obj", "rng")
            for key in keys:
                if key in kwargs:
                    ka[1] = kwargs[key]
                    break
            if count == 1:
                return ka
            keys = ("txt", "text_doc")
            for key in keys:
                if key in kwargs:
                    ka[2] = kwargs[key]
                    break
            return ka

        if not count in (1, 2):
            raise TypeError("get_cursor() got an invalid number of arguments")

        kargs = get_kwargs()

        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        if count == 1:
            return cls._get_cursor_obj(kargs[1])
        txt_doc = mLo.Lo.qi(XTextDocument, kargs[2])
        if txt_doc is None:
            txt = kargs[2]
        else:
            txt = txt_doc.getText()
        return cls._get_cursor_txt(rng=kargs[1], txt=txt)

    @staticmethod
    def _get_cursor_txt(rng: XTextRange, txt: XText) -> XTextCursor:
        try:
            cursor = txt.createTextCursorByRange(rng)
            if cursor is None:
                raise Exception("Cursor is None")
            return cursor
        except Exception as e:
            raise mEx.CursorError(str(e)) from e

    @staticmethod
    def _get_cursor_obj(cursor_obj: DocOrCursor) -> XTextCursor:
        try:
            # https://wiki.openoffice.org/wiki/Writer/API/Text_cursor
            xtxt_cursor = mLo.Lo.qi(XTextCursor, cursor_obj)
            if xtxt_cursor is not None:
                c = xtxt_cursor.getText().createTextCursorByRange(xtxt_cursor)
                if c is None:
                    raise Exception("XTextViewCursor.createTextCursorByRange() result is null")
                return c

            xdoc = mLo.Lo.qi(XTextDocument, cursor_obj)
            if xdoc is not None:
                xtext = xdoc.getText()
                if xtext is None:
                    return None
                c = xtext.createTextCursor()
                if c is None:
                    raise Exception("XTextDocument, XText failed to create text cursor")
                return c
            raise TypeError("cursor_obj is invalid type. Expected XTextDocument or XTextViewCursor")
        except Exception as e:
            raise mEx.CursorError(str(e)) from e
        # XTextViewCursor

    # endregion get_cursor()

    @classmethod
    def get_word_cursor(cls, cursor_obj: DocOrCursor) -> XWordCursor:
        """
        Gets document word cursor

        Args:
            cursor_obj (DocOrCursor): Text Document or Text Cursor

        Raises:
            WordCursorError: If Unable to get cursor

        Returns:
            XWordCursor: Word Cursor
        """
        try:
            if mLo.Lo.is_uno_interfaces(cursor_obj, XTextDocument):
                cursor = cls.get_cursor(cursor_obj)
            else:
                cursor = cursor_obj
            wd_cursor = mLo.Lo.qi(XWordCursor, cursor, True)
            return wd_cursor
        except Exception as e:
            raise mEx.WordCursorError(str(e)) from e

    @classmethod
    def get_sentence_cursor(cls, cursor_obj: DocOrCursor) -> XSentenceCursor:
        """
        Gets document sentence cursor

        Args:
            cursor_obj (DocOrCursor): Text Document or Text Cursor

        Raises:
            SentenceCursorError: If Unable to get cursor

        Returns:
            XSentenceCursor: Sentence Cursor
        """
        try:
            if mLo.Lo.is_uno_interfaces(cursor_obj, XTextDocument):
                cursor = cls.get_cursor(cursor_obj)
            else:
                cursor = cursor_obj
            if cursor is None:
                print("Text cursor is null")
                return None
            sc = mLo.Lo.qi(XSentenceCursor, cursor, True)
            return sc
        except Exception as e:
            raise mEx.SentenceCursorError(str(e)) from e

    @classmethod
    def get_paragraph_cursor(cls, cursor_obj: DocOrCursor) -> XParagraphCursor:
        """
        Gets document paragraph cursor

        Args:
            cursor_obj (DocOrCursor): Text Document or Text Cursor

        Raises:
            ParagraphCursorError: If Unable to get cursor

        Returns:
            XParagraphCursor: Paragraph cursor
        """
        try:
            if mLo.Lo.qi(XTextDocument, cursor_obj) is None:
                cursor = cursor_obj
            else:
                cursor = cls.get_cursor(cursor_obj)
            para_cursor = mLo.Lo.qi(XParagraphCursor, cursor, True)
            return para_cursor
        except Exception as e:
            raise mEx.ParagraphCursorError(str(e)) from e

    @classmethod
    def get_left_cursor(cls, o_sel: XTextRange, o_text: DocOrText) -> XTextCursor:
        """
        Creates a new TextCursor with position left that can travel right

        Args:
            o_sel (XTextRange): Text Range
            o_text (DocOrText): Text document or text.

        Returns:
            XTextCursor: a new instance of a TextCursor which is located at the specified
            TextRange to travel in the given text context.
        """
        doc = mLo.Lo.qi(XTextDocument, o_text)
        if doc is None:
            text = o_text
        else:
            text = doc.getText()

        range_compare = cls.text_range_compare
        if range_compare.compareRegionStarts(o_sel.getEnd(), o_sel) >= 0:
            o_range = o_sel.getEnd()
        else:
            o_range = o_sel.getStart()
        cursor = text.createTextCursorByRange(o_range)
        cursor.goRight(0, False)
        return cursor

    @classmethod
    def get_right_cursor(cls, o_sel: XTextRange, o_text: DocOrText) -> XTextCursor:
        """
        Creates a new TextCursor with position right that can travel left

        Args:
            o_sel (XTextRange): Text Range
            o_text (DocOrText): Text document or text.

        Returns:
            XTextCursor: a new instance of a TextCursor which is located at the specified
            TextRange to travel in the given text context.
        """
        range_compare = cls.text_range_compare
        if range_compare.compareRegionStarts(o_sel.getEnd(), o_sel) >= 0:
            o_range = o_sel.getStart()
        else:
            o_range = o_sel.getEnd()
        cursor = o_text.createTextCursorByRange(o_range)
        cursor.goLeft(0, False)
        return cursor

    @classmethod
    def get_position(cls, cursor: XTextCursor) -> int:
        """
        Gets position of the cursor

        Args:
            cursor (XTextCursor): Text Cursor

        Returns:
            int: Current Cursor Position
        """
        # def get_near_max(l:XTextCursor, r: XTextCursor, jump=10) -> int:
        #     imax = 0
        #     if cls.compare_cursor_ends(l, r) == cls.CompareEnum.BEFORE:
        #         l.goRight(jump, False)
        #         imax = imax + jump
        jmp_amt = 25

        def get_high(l: XTextCursor, r: XTextCursor, jump=jmp_amt, total=0) -> int:
            # OPTIMIZE: get_position.get_high()
            # The idea of this function is to cut down on the number if iterations
            # needed to get the range from cursors left and right positions.
            # Most likely there is an even more efficient way to do this.
            if jump <= 0:
                return 0
            if cls.compare_cursor_ends(l, r) == cls.CompareEnum.BEFORE:
                j = jump + jmp_amt
                l.goRight(j, False)
                return get_high(l, r, j, jump)
            else:
                l.gotoStart(False)
                l.goRight(total, False)
                return total
            # else:
            #     return jump, True

        xcursor = cursor.getText().createTextCursor()
        xcursor.gotoStart(False)
        xcursor.gotoEnd(True)
        xtext = xcursor.getText()
        l_cursor = cls.get_left_cursor(o_sel=xcursor, o_text=xtext)
        r_cursor = cls.get_right_cursor(o_sel=xcursor, o_text=xtext)
        i = 0
        if cls.compare_cursor_ends(c1=l_cursor, c2=r_cursor) < cls.CompareEnum.EQUAL:
            high = get_high(l_cursor, r_cursor)
            # l_cursor.gotoStart(False)

            while cls.compare_cursor_ends(c1=l_cursor, c2=r_cursor) != cls.CompareEnum.EQUAL:
                l_cursor.goRight(1, False)
                i += 1
            i += high
        return i

        # return len(cursor.getText().getString())

    # region ------------- view cursor methods -------------------------

    @classmethod
    def get_text_view_cursor_prop_set(cls, text_doc: XTextDocument) -> XPropertySet:
        """
        Gets properties for document view cursor

        Args:
            text_doc (XTextDocument): Text Document

        Raises:
            MissingInterfaceError: If unable to obtain XPropertySet interface from cursor.

        Returns:
            XPropertySet: Properties
        """
        xview_cursor = cls.get_view_cursor(text_doc)
        props = mLo.Lo.qi(XPropertySet, xview_cursor, True)
        return props

    @staticmethod
    def get_view_cursor(text_doc: XTextDocument) -> XTextViewCursor:
        """
        Gets document view cursor.

        Describes a cursor in a text document's view.

        Args:
            text_doc (XTextDocument): Text Document

        Raises:
            ViewCursorError: If Unable to get cursor

        Returns:
            XTextViewCursor: Text View Cursor

        See Also:
            `LibreOffice API XTextViewCursor <https://api.libreoffice.org/docs/idl/ref/interfacecom_1_1sun_1_1star_1_1text_1_1XTextViewCursor.html>`_
        """

        # https://wiki.openoffice.org/wiki/Writer/API/Text_cursor
        try:
            model = mLo.Lo.qi(XModel, text_doc, True)
            xcontroller = model.getCurrentController()
            supplier = mLo.Lo.qi(XTextViewCursorSupplier, xcontroller, True)
            vc = supplier.getViewCursor()
            if vc is None:
                raise Exception("Supplier return null view cursor")
            return vc
        except Exception as e:
            raise mEx.ViewCursorError(str(e)) from e

    # endregion ---------- view cursor methods -------------------------

    # endregion ---------- model cursor methods ------------------------

    @classproperty
    def text_range_compare(cls) -> XTextRangeCompare:
        """
        Gets text range for comparison operations

        Returns:
            XTextRangeCompare: Text Range Compare instance
        """

        try:
            return cls._text_range_compare
        except AttributeError:
            doc = mLo.Lo.XSCRIPTCONTEXT.getDocument()
            text = doc.getText()
            cls._text_range_compare = mLo.Lo.qi(XTextRangeCompare, text)
        return cls._text_range_compare

    @text_range_compare.setter
    def text_range_compare(cls, value) -> None:
        # raise error on set. Not really necessary but gives feedback.
        raise AttributeError("Attempt to modify read-only class property '%s'." % cls.__name__)

    @staticmethod
    def get_word_count_ooo(text: str, word_type: WordTypeEnum | None = None, locale_lang: str | None = None) -> int:
        """
        Get the number of word in ooo way.

        This method takes into account the current Locale

        Args:
            text (str): string to count the word of
            word_type (WordTypeEnum, optional): type of words to count. Default ``WordTypeEnum.WORD_COUNT``
                Import  line ``from ooodev.utils.selection import WordTypeEnum``
            locale_lang (str, optional): Language such as 'en-US' used to process word boundaries. Defaults to LO's current language.

        Raises:
            CreateInstanceMsfError: If unable to create ``i18n.BreakIterator service``

        Returns:
            int: The number of words
        """
        if word_type is None:
            word_type = WordTypeEnum.WORD_COUNT
        # https://forum.openoffice.org/en/forum/viewtopic.php?f=20&t=82678
        next_wd = Boundary()
        local = Locale()
        # local.Language = "en"
        if locale_lang is None:
            local.Language = mInfo.Info.language
        else:
            local.Language = locale_lang
        num_words = 0
        start_pos = 0
        if word_type > WordTypeEnum.ANY_WORD:
            # intensionally pad the start and end of the string to guarantee we get the first word and clean break on last word
            st = f" {text} "
        else:
            # ANY_WORD
            # no need to pad. All characters will be counted including whitespaces
            st = text

        brk = mLo.Lo.create_instance_mcf(XBreakIterator, "com.sun.star.i18n.BreakIterator")
        if brk is None:
            raise mEx.CreateInstanceMcfError(XBreakIterator, "com.sun.star.i18n.BreakIterator")

        next_wd = brk.nextWord(st, start_pos, local, word_type.value)
        while next_wd.startPos != next_wd.endPos:
            num_words += 1
            nw = next_wd.startPos
            next_wd = brk.nextWord(st, nw, local, word_type.value)

        return num_words

    @classmethod
    def select_next_word(cls, text_doc: XTextDocument) -> bool:
        """
        Select the word right from the current cursor position.

        Args:
            text_doc (XTextDocument): Text Document

        Returns:
            bool: True if go to next word succeeds; Otherwise, False.

        :events:
            .. cssclass:: lo_event

                - :py:attr:`~.events.write_named_event.WriteNamedEvent.WORD_SELECTING` :eventref:`write_word_selecting`
                - :py:attr:`~.events.write_named_event.WriteNamedEvent.WORD_SELECTED` :eventref:`write_word_selected`

        Note:
            The method returning ``True`` does not necessarily mean that the cursor is located at
            the next word, or any word at all! This may happen for example if it travels over empty paragraphs.

        Note:
           Event args ``event_data`` is a dictionary containing ``text_doc``.
        """

        cargs = CancelEventArgs(Selection.select_next_word)
        cargs.event_data = {"text_doc": text_doc}
        _Events().trigger(WriteNamedEvent.WORD_SELECTING, cargs)
        if cargs.cancel:
            return False

        supplier = mLo.Lo.qi(XSelectionSupplier, text_doc.getCurrentController(), True)

        # clear any current selection
        view_curor = cls.get_view_cursor(text_doc=text_doc)
        # view_curor.collapseToEnd()
        view_curor.goRight(0, False)

        # see section 7.5.1 of developers' guide
        index_access = mLo.Lo.qi(XIndexAccess, supplier.getSelection(), True)
        range = mLo.Lo.qi(XTextRange, index_access.getByIndex(0), True)

        # get the XWordCursor and make a selection!
        xText = range.getText()
        word_cursor = cls.get_word_cursor(xText.createTextCursorByRange(range))

        if not word_cursor.isStartOfWord():
            word_cursor.gotoStartOfWord(False)

        result = word_cursor.gotoNextWord(True)
        supplier.select(word_cursor)
        _Events().trigger(WriteNamedEvent.WORD_SELECTED, EventArgs.from_args(cargs))
        return result


def _del_cache_attrs(source: object, e: EventArgs) -> None:
    # clears Write Attributes that are dynamically created
    dattrs = ("_text_range_compare",)
    for attr in dattrs:
        if hasattr(Selection, attr):
            delattr(Selection, attr)


# subscribe to events that warrant clearing cached attribs
_Events().on(LoNamedEvent.RESET, _del_cache_attrs)

__all__ = ("Selection",)



# coding: utf-8
from __future__ import annotations
from typing import cast
from ..utils import lo as mLo

from com.sun.star.awt import XToolkit2
from com.sun.star.awt import XMessageBox

from ooo.dyn.awt.message_box_results import MessageBoxResultsEnum as MessageBoxResultsEnum
from ooo.dyn.awt.message_box_buttons import MessageBoxButtonsEnum as MessageBoxButtonsEnum
from ooo.dyn.awt.message_box_type import MessageBoxType as MessageBoxType


class MsgBox:
    @staticmethod
    def msgbox(
        msg: str,
        title: str = "Message",
        boxtype: MessageBoxType = MessageBoxType.MESSAGEBOX,
        buttons: MessageBoxButtonsEnum | int = MessageBoxButtonsEnum.BUTTONS_OK,
    ) -> MessageBoxResultsEnum:
        """
        Simple message box.

        Args:
            msg (str): the message for display
            title (str, optional): the title of the message box. Defaults to "Message".
            boxtype (MessageBoxType, optional): determines the type of message box to display. Defaults to ``Type.MESSAGEBOX``.
            buttons (MessageBoxButtonsEnum, int, optional): determines what buttons to display. Defaults to ``Buttons.BUTTONS_OK``.

        Returns:
            Results: MessageBoxResultsEnum

            * Button press ``Abort`` return ``MessageBoxResultsEnum.CANCEL``
            * Button press ``Cancel`` return ``MessageBoxResultsEnum.CANCEL``
            * Button press ``Ignore`` returns ``MessageBoxResultsEnum.IGNORE``
            * Button press ``No`` returns ``MessageBoxResultsEnum.NO``
            * Button press ``OK`` returns ``MessageBoxResultsEnum.OK``
            * Button press ``Retry`` returns ``MessageBoxResultsEnum.RETRY``
            * Button press ``Yes`` returns ``MessageBoxResultsEnum.YES``
        """
        if boxtype == MessageBoxType.INFOBOX:
            # this is the default behaviour anyways. So assigning ok to make it official here
            _buttons = MessageBoxButtonsEnum.BUTTONS_OK.value
        else:
            _buttons = buttons

        tk = mLo.Lo.create_instance_mcf(XToolkit2, "com.sun.star.awt.Toolkit")
        parent = tk.getDesktopWindow()
        box = cast(XMessageBox, tk.createMessageBox(parent, boxtype, int(_buttons), str(title), str(msg)))
        return MessageBoxResultsEnum(int(box.execute()))



from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass
from ..validation import check
from ..decorator import enforce
from .base_float_value import BaseFloatValue

if TYPE_CHECKING:
    try:
        from typing import Self
    except ImportError:
        from typing_extensions import Self


@enforce.enforce_types
@dataclass(frozen=True)
class ImageOffset(BaseFloatValue):
    """Represents a Image Offset value between ``0.0`` and ``1.0``"""

    def __post_init__(self) -> None:
        check(
            (self.Value < 0.0 or self.Value >= 1.0) == False,
            f"{self}",
            f"Value of {self.Value} is out of range. Value must be between 0.0 and 1.0",
        )

    def _from_float(self, value: int) -> Self:
        return ImageOffset(value)



# coding: utf-8
from __future__ import annotations
import os
from typing import Any, TYPE_CHECKING
import uno

_DOCS_BUILDING = os.environ.get("DOCS_BUILDING", None) == 'True'
# _DOCS_BUILDING is only true when sphinx is building docs.
# env var DOCS_BUILDING is set in docs/conf.py
_ON_RTD = os.environ.get('READTHEDOCS', None) == 'True'
# env var READTHEDOCS is true when read the docs is building
# maybe not needed as _DOCS_BUILDING is set in conf.py

if _DOCS_BUILDING or _ON_RTD:
    # mock class
    from sphinx.ext.autodoc.mock import _MockObject
    class UnoConst(_MockObject):
        pass
else:
    class UnoConst:
        """
        Get access to Uno Const values without having to directly import them.
        
        This is a singleton class.
        Repeated call to constructor for any given Enum type,
        such as ``com.sun.star.text.ControlCharacter`` will return the same object
        
        Example:
        
            .. code-block:: python
            
                ControlCharacter = UnoConst("com.sun.star.text.ControlCharacter")
                assert ControlCharacter.LINE_BREAK == 1
                
                MyConst = UnoConst("com.sun.star.text.ControlCharacter")
                assert ControlCharacter is MyConst # singleton, same instances

        See Also:
            :py:class:`~.uno_enum.UnoEnum`
        """
        _loaded = {}
        _initialized = False  # This class var is important. It is always False.
                            # The instances will override this with their own, 
                            # set to True.
        def __new__(cls, type_name: str):
            if type_name in cls._loaded:
                return cls._loaded[type_name]
            e = super().__new__(cls)
            e.__dict__["_type_name"] = type_name
            cls._loaded[type_name] = e
            return e
            
        def __init__(self, type_name: str) -> None:
            """
            Constructor
            
            Get access to Uno Const values without having to directly import them.

            Args:
                type_name (str): The namespace of the Const as a string. Such as 'com.sun.star.text.ControlCharacter'
            """
            self._initialized = True
        
        def __getattr__(self, __name: str) -> Any:
            if self._initialized:
                # Provide the caller attributes in whatever ways interest you.
                try:
                    key = __name
                    const = uno.getConstantByName(f"{self._type_name}.{__name}")
                    self.__dict__[key] = const
                    return self.__dict__[key]
                except Exception:
                    raise AttributeError(f"Const {self._type_name} has no attribute {__name}")
            else:
                try:
                    return self.__dict__[__name] # Transparent access to instance vars.
                except KeyError:
                    raise AttributeError(__name)

        def __setattr__(self, key, value):
            if self._initialized:
                pass # Provide caller ways to set attributes in whatever ways.
            else:
                self.__dict__[key] = value # Transparent access.

        @property
        def typeName(self) -> str:
            """
            Gets Uno Type Name

            Returns:
                str: Uno type Name
            """
            return self._type_name

    __all__ = ("UnoConst",)


# coding: utf-8
"""
Draw Named Events.
"""
from __future__ import annotations
from typing import NamedTuple


class DrawNamedEvent(NamedTuple):
    """
    Named events for utils.draw.Draw class
    """

    GET_SHAPES_ERROR = "draw_get_shapes_error"
    """Draw get_shapes error command see :py:meth:`Draw.get_shapes() <.utils.draw.Draw.get_shapes>`"""


# coding: utf-8

# importing using a text file as version source is not supported when packaging scripts using stickytape.
# with open(os.path.join(os.path.dirname(__file__), "VERSION"), "r", encoding="utf-8") as f:
#     version = f.read().strip()

__version__ = "0.6.5"



# coding: utf-8
import os
import sys
from pathlib import Path
import shutil

if sys.platform == "win32":
    import winreg

_INSTALL_PATH = None


def get_soffice_install_path() -> Path:
    """
    Gets the Soffice install path.

    For windows this will be something like: ``C:\Program Files\LibreOffice``.
    For Linux this will be something like: ``/usr/lib/libreoffice``

    Returns:
        Path: install as Path.
    """
    global _INSTALL_PATH
    if _INSTALL_PATH is not None:
        return _INSTALL_PATH
    if sys.platform == "win32":
        # get the path location from registry
        value = ""
        for _key in (
            # LibreOffice 3.4.5,6,7 on Windows
            "SOFTWARE\\LibreOffice\\UNO\\InstallPath",
            # OpenOffice 3.3
            "SOFTWARE\\OpenOffice.org\\UNO\\InstallPath",
        ):
            try:
                value = winreg.QueryValue(winreg.HKEY_LOCAL_MACHINE, _key)
            except Exception as detail:
                value = ""
                _errMess = "%s" % detail
            else:
                break  # first existing key will do
        if value != "":
            _INSTALL_PATH = Path("\\".join(value.split("\\")[:-1]))  # drop the program
            return _INSTALL_PATH

        # failed to get path from registry. Going Manual
        soffice = "soffice.exe"
        p_sf = Path(os.environ["PROGRAMFILES"], "LibreOffice", "program", soffice)
        if p_sf.exists() is False or p_sf.is_file() is False:
            p_sf = Path(os.environ["PROGRAMFILES(X86)"], "LibreOffice", "program", soffice)
        if not p_sf.exists():
            raise FileNotFoundError(f"LibreOffice '{p_sf}' not found.")
        if not p_sf.is_file():
            raise IsADirectoryError(f"LibreOffice '{p_sf}' is not a file.")
        # drop \program\soffice.exe
        # expect C:\Program Files\LibreOffice
        _INSTALL_PATH = p_sf.parent.parent
        return _INSTALL_PATH

    else:
        # unix
        soffice = "soffice"
        # search system path
        s = shutil.which(soffice)
        p_sf = None
        if s is not None:
            # expect '/usr/bin/soffice'
            if os.path.islink(s):
                p_sf = Path(os.path.realpath(s))
            else:
                p_sf = Path(s)
        if p_sf is None:
            s = "/usr/bin/soffice"
            if os.path.islink(s):
                p_sf = Path(os.path.realpath(s))
            else:
                p_sf = Path(s)
        if not p_sf.exists():
            raise FileNotFoundError(f"LibreOffice '{p_sf}' not found.")
        if not p_sf.is_file():
            raise IsADirectoryError(f"LibreOffice '{p_sf}' is not a file.")
        # drop /program/soffice
        _INSTALL_PATH = p_sf.parent.parent
        return _INSTALL_PATH



# coding: utf-8
from __future__ import annotations
from typing import TYPE_CHECKING, Tuple, overload
import uno
from com.sun.star.beans import XPropertySet
from com.sun.star.container import XNameContainer
from com.sun.star.document import XMimeTypeInfo
from com.sun.star.graphic import XGraphicProvider

if TYPE_CHECKING:
    from com.sun.star.graphic import XGraphic
    from com.sun.star.awt import XBitmap

from ooo.dyn.awt.size import Size

from ..utils import lo as mLo
from ..utils import file_io as mFileIO
from ..utils import props as mProps
from ..utils import info as mInfo
from ..exceptions import ex as mEx
from ..utils.type_var import PathOrStr

# see Also: https://ask.libreoffice.org/t/graphicurl-no-longer-works-in-6-1-0-3/35459/3
# see Also: https://tomazvajngerl.blogspot.com/2018/03/improving-image-handling-in-libreoffice.html


class ImagesLo:
    @staticmethod
    def get_bitmap(fnm: PathOrStr) -> XBitmap:
        """
        Gets an image from path

        Args:
            fnm (PathOrStr): path to image

        Raises:
            UnOpenableError: If image is not able to be opened
            Exception: If unable to get bitmap

        Returns:
            XBitmap: Bitmap
        """
        try:
            fp = mFileIO.FileIO.get_absolute_path(fnm)
            bitmap_container = mLo.Lo.create_instance_msf(XNameContainer, "com.sun.star.drawing.BitmapTable")
            if not mFileIO.FileIO.is_openable(fp):
                raise mEx.UnOpenableError(fnm=fp)

            name = str(fp)
            pic_url = mFileIO.FileIO.fnm_to_url(fp)

            # use the filename as the name of the bitmap
            bitmap_container.insertByName(name, pic_url)

            bitmap = bitmap_container.getByName(name)
            # return bitmap.getDIB() # byte sequence
            return bitmap
        except mEx.UnOpenableError:
            raise
        except Exception as e:
            raise Exception(f"Could not create a bitmap container for '{fnm}'") from e

    @staticmethod
    def load_graphic_file(im_fnm: PathOrStr) -> XGraphic:
        """
        Loads a graphic file

        Args:
            im_fnm (PathOrStr): Graphic file path

        Raises:
            FileNotFoundError: If ``im_fnm`` does not exist.
            ImageError: If unable to load graphic

        Returns:
            XGraphic: Graphic
        """
        mLo.Lo.print(f"Loading XGraphic from '{im_fnm}'")
        try:
            _ = mFileIO.FileIO.is_exist_file(fnm=im_fnm, raise_err=True)
            fnm = mFileIO.FileIO.get_absolute_path(im_fnm)
            gprovider = mLo.Lo.create_instance_mcf(
                XGraphicProvider, "com.sun.star.graphic.GraphicProvider", raise_err=True
            )
            file_props = mProps.Props.make_props(URL=mFileIO.FileIO.fnm_to_url(fnm))
            result = gprovider.queryGraphic(file_props)
            if result is None:
                raise mEx.UnKnownError("None Value: queryGraphic() returned None")
            return result
        except FileNotFoundError:
            raise
        except Exception as e:
            raise mEx.ImageError(f"Could not load XGraphic from '{im_fnm}'") from e

    @classmethod
    def get_size_pixels(cls, im_fnm: PathOrStr) -> Size:
        """
        Gets an image/graphic size from image file path

        Args:
            im_fnm (PathOrStr): file path

        Returns:
            Size: Size containing Width and Height
        """
        graphic = cls.load_graphic_file(im_fnm)
        return mProps.Props.get(graphic, "SizePixel")

    @classmethod
    def get_size_100mm(cls, im_fnm: PathOrStr) -> Size:
        """
        The Size of the graphic in ``100th mm``.

        This property may not be available in case of pixel graphics or
        if the logical size can not be determined correctly for some formats
        without loading the whole graphic


        Args:
            im_fnm (PathOrStr): Path to image/graphic

        Raises:
            PropertyError: If Size100thMM property is not available

        Returns:
            Size: Size containing Width and Height
        """
        graphic = cls.load_graphic_file(im_fnm)
        return mProps.Props.get(graphic, "Size100thMM")

    @staticmethod
    def load_graphic_link(graphic_link: object) -> XGraphic:
        """
        Loads a graphic link

        Args:
            graphic_link (object): Object that implements XPropertySet interface

        Raises:
            CreateInstanceMcfError: If unable to create graphic.GraphicProvider instance
            MissingInterfaceError: If graphic_link does not implement XPropertySet interface
            Exception: If unable to load graphic

        Returns:
            XGraphic: Graphic
        """
        xprops = mLo.Lo.qi(XPropertySet, graphic_link, True)

        try:
            graphic = xprops.getPropertyValue("Graphic")
            if graphic is None:
                raise Exception("Grapich is None")
            return graphic
        except Exception as e:
            raise Exception(f"Unable to retrieve graphic") from e

    # region save_graphic()
    @overload
    @staticmethod
    def save_graphic(pic: XGraphic, fnm: PathOrStr) -> None:
        ...

    @overload
    @staticmethod
    def save_graphic(pic: XGraphic, fnm: PathOrStr, im_format: str) -> None:
        ...

    @staticmethod
    def save_graphic(pic: XGraphic, fnm: PathOrStr, im_format: str = "") -> None:
        """
        Save a graphic to disk

        Args:
            pic (XGraphic): Graphic object
            fnm (PathOrStr): File path to save graphic to
            im_format (str, optional): Image format such as ``png``. Defaults to extension of ``fnm``.

        Raises:
            ImageError: If error occurs.
        """
        mLo.Lo.print(f"Saving graphic in '{fnm}'")

        try:
            if pic is None:
                raise TypeError("Expected pic to be XGraphic instance but got None")
            if not im_format:
                im_format = mInfo.Info.get_ext(fnm)
                if not im_format:
                    raise ValueError(
                        "Unable to get image format from fnm. Does fnm have an file extension such as myfile.png?"
                    )
                im_format = im_format.lower()

            gprovider = mLo.Lo.create_instance_mcf(
                XGraphicProvider, "com.sun.star.graphic.GraphicProvider", raise_err=True
            )

            png_props = mProps.Props.make_props(URL=mFileIO.FileIO.fnm_to_url(fnm), MimeType=f"image/{im_format}")

            gprovider.storeGraphic(pic, png_props)
        except Exception as e:
            raise mEx.ImageError(f'Error saving graphic for "{fnm}') from e

    # end region save_graphic()

    @staticmethod
    def get_mime_types() -> Tuple[str, ...]:
        mi = mLo.Lo.create_instance_mcf(XMimeTypeInfo, "com.sun.star.drawing.GraphicExportFilter", raise_err=True)
        return mi.getSupportedMimeTypeNames()

    @classmethod
    def change_to_mime(cls, im_format: str) -> str:
        names = cls.get_mime_types()
        imf = im_format.lower().strip()
        for name in names:
            if imf in name:
                print(f"using mime type: {name}")
                return name
        print("No matching mime type, so using image/png")
        return "image/png"

    @classmethod
    def calc_scale(cls, fnm: PathOrStr, max_width: int, max_height: int) -> Size | None:
        """
        Calculate a new size for the image in fnm that is no bigger than
        maxWidth x maxHeight mm's
        This involves a re-scaling of the image so it is not distorted.
        The new size is returned in mm units
        """
        im_size = cls.get_size_100mm(fnm)  # in 1/100 mm units
        if im_size is None:
            return None
        # calculate the scale factors to obtain these maximums
        width_scale = (max_width * 100) / im_size.Width
        height_scale = (max_height * 100) / im_size.Height

        # use the smallest scale factor
        scale_factor = min(width_scale, height_scale)

        # calculate new dimensions for the image
        w = round(im_size.Width * scale_factor / 100)
        h = round(im_size.Height * scale_factor / 100)
        return Size(w, h)



from enum import Enum
from . import kind_helper


class GalleryKind(str, Enum):
    """
    Gallery Kind.

    Used to encapsulate Gallery names.

    Note:
        Not all gallery name are necessarily lister in this enum.
        Different system and version may have a some different values.

    See Also:
        - :py:meth:`.Gallery.find_gallery_item`
        - :py:meth:`.Gallery.report_galeries`
    """

    BPMN = "BPMN"
    FLOW_CHART = "Flow chart"
    SOUNDS = "Sounds"
    ARROWS = "Arrows"
    DIAGRAMS = "Diagrams"
    NETWORK = "Network"
    SHAPES = "Shapes"
    BULLETS = "Bullets"
    ICONS = "Icons"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_str(s: str) -> "GalleryKind":
        """
        Gets an ``GalleryKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``GalleryKind`` instance.

        Returns:
            GalleryKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, GalleryKind)



from enum import Enum
from . import kind_helper


class CurveKind(Enum):
    """
    Curve Kind Enum. Regression line types.

    Usage:

    .. code-block:: python

        # compare enums
        >>> x = CurveKind.MOVING_AVERAGE
        >>> y = CurveKind.MOVING_AVERAGE
        >>> assert x == y

        >>> print(x.curve)
        5
        >>> print(x.label)
        Moving average
        >>> print(x.name)
        MOVING_AVERAGE
        >>> print(CurveKind.LINEAR.to_namespace())
        com.sun.star.chart2.LinearRegressionCurve
    """

    LINEAR = 0, "Linear", "LinearRegressionCurve"
    LOGARITHMIC = 1, "Logarithmic", "LogarithmicRegressionCurve"
    EXPONENTIAL = 2, "Exponential", "ExponentialRegressionCurve"
    POWER = 3, "Power", "PotentialRegressionCurve"
    POLYNOMIAL = 4, "Polynomial", "PolynomialRegressionCurve"
    MOVING_AVERAGE = 5, "Moving average", "MovingAverageRegressionCurve"

    def __init__(self, curve: int, label: str, ns: str):
        self.curve = curve
        self.label = label
        self.ns_regression_curve = ns

    def to_namespace(self) -> str:
        """
        Gets the full UNO namespace value of CurveKind instance.

        Returns:
            str: String namespace such as ``com.sun.star.chart2.LinearRegressionCurve``
        """
        return f"com.sun.star.chart2.{self.ns_regression_curve}"

    @staticmethod
    def from_str(s: str) -> "CurveKind":
        """
        Gets an ``CurveKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``CurveKind`` instance.

        Returns:
            CurveKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, CurveKind)



from ooodev.mock import mock_g
from ..mock import mock_g

try:
    import tkinter as tk
except ImportError:
    if mock_g.DOCS_BUILDING:
        # ignoring simply because windows and sphinx don't see tkinter when python interpreter is pointed to LibreOffice python.exe
        pass
    else:
        raise


class Window:
    """Creates a dialog Window"""

    def __init__(self, title: str = "Input", input_msg: str = "Enter Input:", is_password: bool = False) -> None:
        """
        Class Constructor

        Args:
            title (str, optional): Title to display for Dialog. Defaults to "Input".
            input_msg (_type_, optional): Message to display for Dialog. Defaults to "Enter Input:".
            is_password (bool, optional): Determines if input box is masked for password input. Defaults to False.
        """
        self.root = tk.Tk()
        self.root.geometry("300x70")
        self.root.title(title)
        self.parent = tk.Frame(self.root, padx=10, pady=10)
        self.parent.pack(fill=tk.BOTH, expand=True)
        mask = "*" if is_password else ""
        self.password = tk.StringVar()  # Password variable
        self.passEntry = self._make_entry(self.parent, input_msg, 16, show=mask, textvariable=self.password)
        submit = tk.Button(self.root, text="OK", command=self.get_input)

        # passEntry.pack(pady=12, padx=8)
        submit.pack()

        self.root.mainloop()

    def get_input(self) -> str:
        """
        Gets the input from the dialog

        Returns:
            str: Dialog Input or empty string.
        """
        p = self.password.get()  # get password from entry
        self.root.quit()
        # print("Password:", p)
        return p

    def _make_entry(self, parent, caption, width=None, **options):
        tk.Label(parent, text=caption).pack(side=tk.TOP)
        entry = tk.Entry(parent, **options)
        if width:
            entry.config(width=width)
        entry.pack(side=tk.TOP, padx=10, fill=tk.BOTH)
        return entry



from __future__ import annotations
from abc import abstractmethod
from typing import TYPE_CHECKING
from dataclasses import dataclass
import math

if TYPE_CHECKING:
    try:
        from typing import Self
    except ImportError:
        from typing_extensions import Self

# Note that from __future__ import annotations converts annotations to string.
# this means that @enforce.enforce_types will see string as type. This is fine in
# most cases. Especially for built in types.


@dataclass(frozen=True)
class BaseFloatValue:
    """Base class for Flaot Value"""

    Value: float
    """Float value."""

    @abstractmethod
    def _from_float(self, value: float) -> Self:
        ...

    # Override "int" method
    def __float__(self) -> float:
        return self.Value

    def __int__(self) -> float:
        return int(self.Value)

    def __add__(self, other: object) -> Self:
        try:
            i = float(other)
            return self._from_float(self.Value + i)
        except AssertionError:
            raise
        except Exception:
            return NotImplemented

    def __radd__(self, other: object) -> Self:
        # angle = sum([ang1, ang2, ang3])
        # will result in TypeError becuase sum() start with 0
        # this will force a call to __radd__
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __eq__(self, other: object) -> bool:
        # By default, __ne__() delegates to __eq__() and inverts the result unless it is NotImplemented.
        # There are no other implied relationships among the comparison operators,
        # for example, the truth of (x<y or x==y) does not imply x<=y.
        try:
            i = float(other)
            return math.isclose(i, self.Value)
        except Exception as e:
            return False

    def __sub__(self, other: object) -> Self:
        try:
            i = float(other)
            return self._from_float(self.Value - i)
        except AssertionError:
            raise
        except Exception:
            return NotImplemented

    def __rsub__(self, other: object) -> Self:
        try:
            i = float(other)
            return self._from_float(i - self.Value)
        except AssertionError:
            raise
        except Exception:
            return NotImplemented

    def __mul__(self, other: object) -> Self:
        try:
            i = float(other)
            return self._from_float(self.Value * i)
        except AssertionError:
            raise
        except Exception:
            return NotImplemented

    def __rmul__(self, other: int) -> Self:
        if other == 0:
            return self
        else:
            return self.__mul__(other)

    def __lt__(self, other: object) -> bool:
        try:
            i = float(other)
            if math.isclose(i, self.Value):
                return False
            return self.Value < i
        except Exception:
            return NotImplemented

    def __le__(self, other: object) -> bool:
        try:
            i = float(other)
            if math.isclose(i, self.Value):
                return True
            return self.Value <= i
        except Exception:
            return NotImplemented

    def __gt__(self, other: object) -> bool:
        try:
            i = float(other)
            if math.isclose(i, self.Value):
                return False
            return self.Value > i
        except Exception:
            return NotImplemented

    def __ge__(self, other: object) -> bool:
        try:
            i = float(other)
            if math.isclose(i, self.Value):
                return True
            return self.Value >= i
        except Exception:
            return NotImplemented

    def __abs__(self) -> int:
        return abs(self.Value)



# coding: utf-8
from typing import Any, Callable, TypeVar, Generic

# import functools

T = TypeVar("T")


class classproperty(Generic[T]):
    def __init__(self, fget: Callable[[Any], T], fset=None) -> None:
        # https://stackoverflow.com/questions/71018821/property-decorator-type-hints
        self.fget = fget
        self.fset = fset
        # https://stackoverflow.com/a/17705456/1171746
        # no need to update wrapper in this case.
        # becasue this is a decorater in a metaclass
        # i am not sure yet if the doc string is accessible at all
        # functools.update_wrapper(self, fget)  ## TA-DA! ##

    def __get__(self, obj, objtype=None) -> T:
        val = self.fget(obj)
        return val

    def __set__(self, obj, value: T) -> T:
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        self.fset = func
        return self


class classinstanceproperty(Generic[T]):
    def __init__(self, fget: Callable[[Any], T], fset=None) -> None:
        self.fget = fget
        self.fset = fset
        # functools.update_wrapper(self, fget)  ## TA-DA! ##

    def __get__(self, obj, objtype=None) -> T:
        return self.fget(obj)

    def __set__(self, obj, value: T) -> T:
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        self.fset = func
        return self


class StaticProperty(type):
    def __new__(self, name, bases, props):
        class_properties = {}
        to_remove = {}
        for key, value in props.items():
            if isinstance(value, (classproperty, classinstanceproperty)):
                class_properties[key] = value
                if isinstance(value, classproperty):
                    to_remove[key] = value

        for key in to_remove:
            props.pop(key)

        HoistMeta = type("HoistMeta", (type,), class_properties)
        return HoistMeta(name, bases, props)



from enum import Enum
from . import kind_helper

# these are helper lookups
# Example Usage:
#   Draw.set_style(shape=conn_shape, graphic_styles=g_styles, style_name=GraphicStyleKind.ARROW_LINE)
#
# see the Grouper Sample.
# value can be obtaind in the follow mannor.
# g_styles = Info.get_style_container(doc=doc, family_style_name="graphics")
# Info.show_container_names("Graphic styles", g_styles)


class GraphicStyleKind(str, Enum):
    """
    Graphic Styles
    """

    STANDARD = "standard"
    """The style Default (standard) is used for newly inserted filled rectangles, filled ellipses, lines, connectors, text boxes, and 3D objects."""
    OBJECT_WITHOUT_FILL = "objectwithoutfill"
    OBJECT_WITH_NO_FILL_AND_NO_LINE = "Object with no fill and no line"
    TEXT = "Text"
    A4 = "A4"
    TITLE_A4 = "Title A4"
    HEADING_A4 = "Heading A4"
    TEXT_A4 = "Text A4"
    TITLE_A0 = "Title A0"
    HEADING_A0 = "Heading A0"
    TEXT_A0 = "Text A0"
    GRAPHIC = "Graphic"
    SHAPES = "Shapes"
    FILLED = "Filled"
    FILLED_BLUE = "Filled Blue"
    FILLED_GREEN = "Filled Green"
    FILLED_RED = "Filled Red"
    FILLED_YELLOW = "Filled Yellow"
    OUTLINED = "Outlined"
    OUTLINED_BLUE = "Outlined Blue"
    OUTLINED_GREEN = "Outlined Green"
    OUTLINED_RED = "Outlined Red"
    OUTLINED_YELLOW = "Outlined Yellow"
    LINES = "Lines"
    ARROW_LINE = "Arrow Line"
    ARROW_DASHED = "Arrow Dashed"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_str(s: str) -> "GraphicStyleKind":
        """
        Gets an ``GraphicStyleKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``GraphicStyleKind`` instance.

        Returns:
            GraphicStyleKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, GraphicStyleKind)



from enum import Enum
from . import kind_helper


class DrawingHatchingKind(str, Enum):
    """
    Drawing (some) Hatching Values.
    The actual value may change from version to version and may be different
    in other languages.

    These are just quick suggestion that show up in Draw Hatching options.
    """

    BLACK_0_DEGREES = "Black 0 Degrees"
    BLACK_180_DEGREES_CROSSED = "Black 180 Degrees Crossed"
    BLACK_90_DEGREES = "Black 90 Degrees"
    BLUE_45_DEGREES = "Blue 45 Degrees"
    BLUE_45_DEGREES_CROSSED = "Blue 45 Degrees Crossed"
    BLUE_NEG_45_DEGREES = "Blue -45 Degrees"
    GREEN_30_DEGREES = "Green 30 Degrees"
    GREEN_60_DEGREES = "Green 60 Degrees"
    GREEN_90_DEGREES_TRIPLE = "Green 90 Degrees Triple"
    RED_45_DEGREES = "Red 45 Degrees"
    RED_90_DEGREES_CROSSED = "Red 90 Degrees Crossed"
    RED_NEG_45_DEGREES_TRIPLE = "Red -45 Degrees Triple"
    YELLOW_45_DEGREES = "Yellow 45 Degrees"
    YELLOW_45_DEGREES_CROSSED = "Yellow 45 Degrees Crossed"
    YELLOW_45_DEGREES_TRIPLE = "Yellow 45 Degrees Triple"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_str(s: str) -> "DrawingHatchingKind":
        """
        Gets an ``DrawingHatchingKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``DrawingHatchingKind`` instance.

        Returns:
            DrawingHatchingKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, DrawingHatchingKind)



from enum import Enum
from . import kind_helper


class DrawingNameSpaceKind(str, Enum):
    """
    Draw Namespace kind.

    See Also:
        :py:meth:`.Draw.find_shape_by_type`
    """

    BULLETS_TEXT = "com.sun.star.presentation.OutlinerShape"
    SHAPE_TYPE_FOOTER = "com.sun.star.presentation.FooterShape"
    SHAPE_TYPE_NOTES = "com.sun.star.presentation.NotesShape"
    SHAPE_TYPE_PAGE = "com.sun.star.presentation.PageShape"
    SUBTITLE_TEXT = "com.sun.star.presentation.SubtitleShape"
    TITLE_TEXT = "com.sun.star.presentation.TitleTextShape"
    OLE2_SHAPE = "com.sun.star.drawing.OLE2Shape"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_str(s: str) -> "DrawingNameSpaceKind":
        """
        Gets an ``DrawingNameSpaceKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``DrawingNameSpaceKind`` instance.

        Returns:
            DrawingNameSpaceKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, DrawingNameSpaceKind)



# coding: utf-8
from __future__ import annotations
from enum import Enum, IntEnum, IntFlag
from typing import Any, Type
from ...utils import gen_util as mGenUtil


def enum_class_new(cls, value: Any) -> Enum:
    """
    New (__new__) method for dynamically created enum classes

    Args:
        value (object): Can be Enum, Enum.value, str

    Raises:
        ValueError: if unable to match enum instance

    Returns:
        [Enum]: Enum Instance

    Example:
        ..code-block:: python

            >>> e = DrawingHatchingKind("YELLOW_45_DEGREES")
            >>> print(e.value)
            Yellow 45 Degrees
            >>> e = DrawingHatchingKind(DrawingHatchingKind.YELLOW_45_DEGREES)
            >>> print(e.value)
            Yellow 45 Degrees
            >>> e = DrawingHatchingKind(DrawingHatchingKind.YELLOW_45_DEGREES.value)
            >>> print(e.value)
            Yellow 45 Degrees
    """
    if isinstance(value, str):
        if hasattr(cls, value):
            return getattr(cls, value)
    _type = type(value)
    if _type is cls:
        return value
    raise ValueError("%r is not a valid %s" % (value, cls.__name__))


def enum_from_string(s: str, ec: Type[Enum]) -> Enum:
    """
    Gets an enum instance from a string

    Args:
        s (str): Name of enum instance.
            ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
            ``hypen-case``, ``normal case``.
        ec (Enum): Enum to get Enum instance from

    Raises:
        ValueError: If ``s`` string is empty.
        AttributeError: If unable to get enum instance.

    Returns:
        Enum: Enum Instance
    """
    if not s:
        raise ValueError("from_str arg s cannot be an empty value")

    try:
        return getattr(ec, s.upper())
    except AttributeError:
        pass

    if issubclass(ec, (IntEnum, IntFlag)):
        try:
            return ec(int(s))
        except ValueError:
            pass
        if s.upper().startswith("0X"):
            try:
                return ec(int(s, 16))
            except ValueError:
                pass

    e_str = mGenUtil.Util.to_single_space(s).replace("-", "_").replace(" ", "_")
    if "_" in e_str:
        e_str = e_str.upper()
    else:
        # handle pascalCase and CamelCase words
        e_str = mGenUtil.Util.to_snake_case_upper(e_str)
    return getattr(ec, e_str)



from __future__ import annotations
from typing import TYPE_CHECKING
from ..mock import mock_g

if mock_g.DOCS_BUILDING:
    from ..mock import unohelper
else:
    import unohelper

from com.sun.star.frame import XTerminateListener

if TYPE_CHECKING:
    from com.sun.star.lang import EventObject


class XTerminateAdapter(unohelper.Base, XTerminateListener):
    """
    XTerminateListener implementation

    See Also:
        `API XTerminateListener <https://api.libreoffice.org/docs/idl/ref/interfacecom_1_1sun_1_1star_1_1frame_1_1XTerminateListener.html>`_
    """

    def notifyTermination(self, Event: EventObject) -> None:
        """
        is called when the master environment is finally terminated.

        No veto will be accepted then.
        """
        pass

    def queryTermination(self, Event: EventObject) -> None:
        """
        is called when the master environment (e.g., desktop) is about to terminate.

        Termination can be intercepted by throwing TerminationVetoException. Interceptor will be the new owner of desktop and should call XDesktop.terminate() after finishing his own operations.

        Raises:
            TerminationVetoException: ``TerminationVetoException``
        """
        pass

    def disposing(self, event: EventObject) -> None:
        """
        gets called when the broadcaster is about to be disposed.

        All listeners and all other objects, which reference the broadcaster
        should release the reference to the source. No method should be invoked
        anymore on this object ( including XComponent.removeEventListener() ).

        This method is called for every listener registration of derived listener
        interfaced, not only for registrations at XComponent.
        """
        # from com.sun.star.lang.XEventListener
        pass



# coding: utf-8
"""
Various color conversions utilities.
"""
from __future__ import annotations
import math
import colorsys
from typing import Union, NamedTuple, overload, NewType
import numbers

from ..utils import gen_util as mGenUtil

# ref: https://gist.github.com/mathebox/e0805f72e7db3269ec22

MAX_COLOR = 255
"""Max Color Value"""
MIN_COLOR = 0
"""Min Color Value"""

Color = NewType("Color", int)
"""Color Type. Int RGB Value"""


class CommonColor:
    """
    Named Colors.

    See Also:
        - `Wikipedia Web colors <https://en.wikipedia.org/wiki/Web_colors>`_
        - `Hex Color Chart <https://www.quackit.com/css/color/charts/hex_color_chart.cfm>`_

    Example:

        .. code-block:: python

            def set_chart(sheet: XSpreadsheet, range_addr: CellRangeAddress) -> None:
                chart = Chart2.insert_chart(sheet, range_addr, "A22", 20, 11, Chart2.ChartLookup.Column.TEMPLATE_PERCENT.COLUMN_DEEP_3D)
                Chart2.set_background_colors(chart, CommonColor.LIGHT_GREEN, CommonColor.BROWN)
    """

    # https://en.wikipedia.org/wiki/Web_colors
    # https://www.quackit.com/css/color/charts/hex_color_chart.cfm
    # some hex values for commonly used colors
    # Pink colors
    DEEP_PINK = Color(0xFF1493)
    HOT_PINK = Color(0xFF69B4)
    LIGHT_PINK = Color(0xFFB6C1)
    MEDIUM_VIOLET_RED = Color(0xC71585)
    PALE_VIOLET_RED = Color(0xDB7093)
    PINK = Color(0xFFC0CB)

    # red colors
    CRIMSON = Color(0xDC143C)
    DARK_RED = Color(0x8B0000)
    DARK_SALMON = Color(0xE9967A)
    FIRE_BRICK = Color(0xB22222)
    INDIAN_RED = Color(0xCD5C5C)
    LIGHT_CORAL = Color(0xF08080)
    LIGHT_SALMON = Color(0xFFA07A)
    RED = Color(0xFF0000)
    SALMON = Color(0xFA8072)

    # Oranges
    CORAL = Color(0xFF7F50)
    DARK_ORANGE = Color(0xFF8C00)
    ORANGE = Color(0xFFA500)
    ORANGE_RED = Color(0xFF4500)
    TOMATO = Color(0xFF6347)

    # Yellow colors
    DARK_KHAKI = Color(0xBDB76B)
    GOLD = Color(0xFFD700)
    KHAKI = Color(0xF0E68C)
    LEMON_CHIFFON = Color(0xFFFACD)
    LIGHT_GOLDENROD_YELLOW = Color(0xFAFAD2)
    LIGHT_YELLOW = Color(0xFFFFE0)
    MOCCASIN = Color(0xFFE4B5)
    PALE_GOLDENROD = Color(0xEEE8AA)
    PAPAYA_WHIP = Color(0xFFEFD5)
    PEACH_PUFF = Color(0xFFDAB9)
    YELLOW = Color(0xFFFF00)

    # Purples
    BLUE_VIOLET = Color(0x8A2BE2)
    DARK_MAGENTA = Color(0x8B008B)
    DARK_ORCHID = Color(0x9932CC)
    DARK_SLATE_BLUE = Color(0x483D8B)
    DARK_VIOLET = Color(0x9400D3)
    FUCHSIA = Color(0xFF00FF)
    INDIGO = Color(0x4B0082)
    LAVENDER = Color(0xE6E6FA)
    MAGENTA = Color(0xFF00FF)
    MEDIUM_ORCHID = Color(0xBA55D3)
    MEDIUM_PURPLE = Color(0x9370DB)
    MEDIUM_SLATE_BLUE = Color(0x7B68EE)
    ORCHID = Color(0xDA70D6)
    PLUM = Color(0xDDA0DD)
    PURPLE = Color(0x800080)
    REBECCA_PURPLE = Color(0x663399)
    SLATE_BLUE = Color(0x6A5ACD)
    THISTLE = Color(0xD8BFD8)
    VIOLET = Color(0xEE82EE)

    # Greens
    CHARTREUSE = Color(0x7FFF00)
    DARK_CYAN = Color(0x008B8B)
    DARK_GREEN = Color(0x006400)
    DARK_OLIVE_GREEN = Color(0x556B2F)
    DARK_SEA_GREEN = Color(0x8FBC8F)
    FOREST_GREEN = Color(0x228B22)
    GREEN = Color(0x008000)
    GREEN_YELLOW = Color(0xADFF2F)
    LAWN_GREEN = Color(0x7CFC00)
    LIGHT_GREEN = Color(0x90EE90)
    LIGHT_SEA_GREEN = Color(0x20B2AA)
    LIME = Color(0x00FF00)
    LIME_GREEN = Color(0x32CD32)
    MEDIUM_AQUAMARINE = Color(0x66CDAA)
    MEDIUM_SEA_GREEN = Color(0x3CB371)
    MEDIUM_SPRING_GREEN = Color(0x00FA9A)
    OLIVE = Color(0x808000)
    OLIVE_DRAB = Color(0x6B8E23)
    PALE_GREEN = Color(0x98FB98)
    SEA_GREEN = Color(0x2E8B57)
    SPRING_GREEN = Color(0x00FF7F)
    TEAL = Color(0x008080)
    YELLOW_GREEN = Color(0x9ACD32)

    # Blues/Cyans
    AQUA = Color(0x00FFFF)
    AQUAMARINE = Color(0x7FFFD4)
    BLUE = Color(0x0000FF)
    CADET_BLUE = Color(0x5F9EA0)
    CORNFLOWER_BLUE = Color(0x6495ED)
    CYAN = Color(0x00FFFF)
    DARK_BLUE = Color(0x00008B)
    DARK_TURQUOISE = Color(0x00CED1)
    DEEP_SKY_BLUE = Color(0x00BFFF)
    DODGER_BLUE = Color(0x1E90FF)
    LIGHT_BLUE = Color(0xADD8E6)
    LIGHT_CYAN = Color(0xE0FFFF)
    LIGHT_SKY_BLUE = Color(0x87CEFA)
    LIGHT_STEEL_BLUE = Color(0xB0C4DE)
    MEDIUM_BLUE = Color(0x0000CD)
    MEDIUM_TURQUOISE = Color(0x48D1CC)
    MIDNIGHT_BLUE = Color(0x191970)
    NAVY = Color(0x000080)
    PALE_TURQUOISE = Color(0xAFEEEE)
    POWDER_BLUE = Color(0xB0E0E6)
    ROYAL_BLUE = Color(0x4169E1)
    SKY_BLUE = Color(0x87CEEB)
    STEEL_BLUE = Color(0x4682B4)
    TURQUOISE = Color(0x40E0D0)

    # Browns
    BISQUE = Color(0xFFE4C4)
    BLANCHED_ALMOND = Color(0xFFEBCD)
    BROWN = Color(0xA52A2A)
    BURLY_WOOD = Color(0xDEB887)
    CHOCOLATE = Color(0xD2691E)
    CORNSILK = Color(0xFFF8DC)
    DARK_GOLDENROD = Color(0xB8860B)
    GOLDENROD = Color(0xDAA520)
    MAROON = Color(0x800000)
    NAVAJO_WHITE = Color(0xFFDEAD)
    PERU = Color(0xCD853F)
    ROSY_BROWN = Color(0xBC8F8F)
    SADDLE_BROWN = Color(0x8B4513)
    SANDY_BROWN = Color(0xF4A460)
    SIENNA = Color(0xA0522D)
    TAN = Color(0xD2B48C)
    WHEAT = Color(0xF5DEB3)

    # Whites
    ALICE_BLUE = Color(0xF0F8FF)
    ANTIQUE_WHITE = Color(0xFAEBD7)
    AZURE = Color(0xF0FFFF)
    BEIGE = Color(0xF5F5DC)
    FLORAL_WHITE = Color(0xFFFAF0)
    GHOST_WHITE = Color(0xF8F8FF)
    HONEYDEW = Color(0xF0FFF0)
    IVORY = Color(0xFFFFF0)
    LAVENDER_BLUSH = Color(0xFFF0F5)
    LINEN = Color(0xFAF0E6)
    MINT_CREAM = Color(0xF5FFFA)
    MISTY_ROSE = Color(0xFFE4E1)
    OLD_LACE = Color(0xFDF5E6)
    SEASHELL = Color(0xFFF5EE)
    SNOW = Color(0xFFFAFA)
    WHITE = Color(0xFFFFFF)
    WHITE_SMOKE = Color(0xF5F5F5)

    # Greys
    BLACK = Color(0x000000)
    DARK_GRAY = Color(0xA9A9A9)
    DARK_GREY = Color(0xA9A9A9)
    DARK_SLATE_GRAY = Color(0x2F4F4F)
    DARK_SLATE_GREY = Color(0x2F4F4F)
    DIM_GRAY = Color(0x696969)
    DIM_GREY = Color(0x696969)
    GAINSBORO = Color(0xDCDCDC)
    GRAY = Color(0x808080)
    GREY = Color(0x808080)
    LIGHT_GRAY = Color(0xD3D3D3)
    LIGHT_GREY = Color(0xD3D3D3)
    LIGHT_SLATE_GRAY = Color(0x778899)
    LIGHT_SLATE_GREY = Color(0x778899)
    SILVER = Color(0xC0C0C0)
    SLATE_GRAY = Color(0x708090)
    SLATE_GREY = Color(0x708090)

    # other
    PALE_BLUE = Color(0xD6EBFF)

    @classmethod
    def from_str(cls, str_color: str) -> Color:
        """
        Convert string value to a Color value

        ``str_color`` can be a hex value, a integer value as string or any named value in ``CommonColor``

        ``str_color`` can contain single spaces or ``_`` or ``-``.
        ``MEDIUM_SEA_GREEN``, ``MEDIUM-SEA-GREEN``, ``MEDIUM SEA GREEN``, ``medium sea green``, ``mediumSeaGreen`` , ``MediumSeaGreen`` are
        all equivalent.

        Args:
            str_color (str): Value to convert. Case insensitive

        Raises:
            ValueError: If unable to convert.

        Returns:
            Color: Convert value as Color
        """
        if not str_color:
            raise ValueError("str_color contains no value to convert to Color")
        if str_color.isalnum():
            try:
                # try base 10
                i = int(str_color)
            except ValueError:
                # try hex
                try:
                    i = int(str_color, 16)
                except ValueError:
                    pass
                else:
                    return Color(abs(i))
            else:
                return Color(abs(i))
        c_str = mGenUtil.Util.to_single_space(str_color).replace("-", "_").replace(" ", "_")
        if "_" in c_str:
            c_str = c_str.upper()
        else:
            # handle pascalCase and CamelCase words
            c_str = mGenUtil.Util.to_snake_case_upper(c_str)
        try:
            return Color(int(getattr(cls, c_str)))
        except AttributeError:
            raise ValueError("str_color is not a valid color")


class RGB(NamedTuple):
    red: int
    """Red color as int"""
    green: int
    """Green color as int"""
    blue: int
    """Blue color as int"""

    def to_int(self) -> int:
        """
        Gets instance as rgb int

        Returns:
            int: red, green, blue encoded as int.
        """
        return rgb_to_int(self)

    def to_color(self) -> Color:
        """
        Gets instance as rgb Color

        Returns:
            Color: red, green, blue encoded as Color.
        """
        return Color(self.to_int())

    def to_hex(self) -> str:
        """
        Gets instance as hex string in format of ``ff3322``

        Returns:
            str: red, green, blue encoded as hex string.
        """
        return rgb_to_hex(self)

    def isvalid(self) -> bool:
        """
        Gets if the value of red, green and blue are valid.

        Returns:
            bool: ``True`` if valid; Otherwise, ``False``
        """
        result = True
        for i in self:
            if i < MIN_COLOR or i > MAX_COLOR:
                result = False
                break
        return result

    @staticmethod
    def from_int(rgb_int: int) -> "RGB":
        """
        Gets a color instance from int that represents a rgb color.

        Args:
            rgb_int (int): int that contains rgb color data.

        Returns:
            RGB: Color information as RGB struct.
        """
        return int_to_rgb(rgb_int=rgb_int)

    @classmethod
    def from_color(cls, c: Color) -> "RGB":
        """
        Gets a color instance from input color that represents a rgb color.

        Args:
            c (Color): Color that contains rgb color data.

        Returns:
            RGB: Color information as RGB struct.
        """
        return cls.from_int(int(c))

    @staticmethod
    def from_hex(rgb_hex: str) -> "RGB":
        """
        Gets a color instance from int that represents a rgb color.

        Args:
            rgb_int (int): int that contains rgb color data.

        Returns:
            RGB: Color information as RGB struct.
        """
        return int_to_rgb(rgb_int=int(rgb_hex, 16))

    def get_luminance(self) -> float:
        """
        Gets luminance value for current color

        Returns:
            float: luminance value
        """
        # http://www.w3.org/TR/2008/REC-WCAG20-20081211/#relativeluminancedef

        RsRGB = self.red / 255
        GsRGB = self.green / 255
        BsRGB = self.blue / 255

        if RsRGB <= 0.03928:
            R = RsRGB / 12.92
        else:
            R = math.pow(((RsRGB + 0.055) / 1.055), 2.4)
        if GsRGB <= 0.03928:
            G = GsRGB / 12.92
        else:
            G = math.pow(((GsRGB + 0.055) / 1.055), 2.4)
        if BsRGB <= 0.03928:
            B = BsRGB / 12.92
        else:
            B = math.pow(((BsRGB + 0.055) / 1.055), 2.4)
        return (0.2126 * R) + (0.7152 * G) + (0.0722 * B)

    def get_brightness(self) -> int:
        """
        Gets brightness from 0 (dark) to 255 (light)

        Returns:
            int: brightness level
        """
        # http://www.w3.org/TR/AERT#color-contrast
        return round(((self.red * 299) + (self.green * 587) + (self.blue * 114)) / 1000)

    def is_dark(self) -> bool:
        """
        Get is current color is dark.
        If color has a brightness less than ``128`` it is considered dark.

        Returns:
            bool: True if color is dark; Otherwise, False
        """
        return self.get_brightness() < 128

    def is_light(self) -> bool:
        """
        Get is current color is light.
        If color has a brightness Greater than ``128`` it is considered light.

        Returns:
            bool: True if color is light; Otherwise, False
        """
        return not self.is_dark()

    def __str__(self) -> str:
        return "rgb(" + round(self.red) + ", " + round(self.green) + ", " + round(self.blue) + ")"


class HSL(NamedTuple):
    hue: float
    saturation: float
    lightness: float

    def __str__(self) -> str:
        return "hls(" + f"{self.hue:.6f}" + ", " + f"{self.saturation:.6f}" + ", " + f"{self.lightness:.6f}" + ")"


class HSV(NamedTuple):
    hue: float
    saturation: float
    value: float

    def __str__(self) -> str:
        return "hlv(" + f"{self.hue:.6f}" + ", " + f"{self.saturation:.6f}" + ", " + f"{self.value:.6f}" + ")"


def clamp(value: float, min_value: float, max_value: float) -> float:
    """
    Constrains a value to a min and an max value

    Args:
        value (float): Value to constrain
        min_value (float): Min allowed value
        max_value (float): Max allowed value

    Returns:
        float: constrained value if value is outside of min_value or max_value; Otherwise, value.
    """
    return max(min_value, min(max_value, value))


def clamp01(value: float) -> float:
    """
    Gets a value that is constrained between 0.0 and 1.0

    Args:
        value (float): Value

    Returns:
        float: A value that is no less than 0.0 and no greater then 1.0
    """
    return clamp(value, 0.0, 1.0)


def hue_to_rgb(h: float) -> RGB:
    """
    Converts a hue to instance of red, green, blue

    Args:
        h (float): hue to convert

    Returns:
        rgb: instance containing red, green, blue
    """
    r = abs(h * 6.0 - 3.0) - 1.0
    g = 2.0 - abs(h * 6.0 - 2.0)
    b = 2.0 - abs(h * 6.0 - 4.0)
    return RGB(red=round(clamp01(r)), green=round(clamp01(g)), blue=round(clamp01(b)))


def hsl_to_rgb(c: HSL) -> RGB:
    """
    Converts hue, saturation, lightness to red, green, blue

    Args:
        c (hsv): instance containing hue, saturation, lightness

    Returns:
        rgb: instance containing red, green, blue
    """
    h = c.hue
    l = c.lightness
    s = c.saturation

    t = colorsys.hls_to_rgb(h=h, l=l, s=s)
    return RGB(
        red=round(t[0] * MAX_COLOR),
        green=round(t[1] * MAX_COLOR),
        blue=round(t[2] * MAX_COLOR),
    )


def rgb_to_hsv(c: RGB) -> HSV:
    """
    Converts red, green, blue to hue, saturation, value

    Args:
        c (RGB): instance containing red, green, blue

    Returns:
        HSV: instance containing hue, saturation, value
    """
    r = float(c.red / MAX_COLOR)
    g = float(c.green / MAX_COLOR)
    b = float(c.blue / MAX_COLOR)
    t = colorsys.rgb_to_hsv(r=r, g=g, b=b)
    return HSV(hue=t[0], saturation=t[1], value=t[2])


def hsv_to_rgb(c: HSV) -> RGB:
    """
    Converts hue, saturation, value to red, green, blue

    Args:
        c (hsv): instance containing hue, saturation, value

    Returns:
        rgb: instance containing red, green, blue
    """
    h = c.hue
    s = c.saturation
    v = c.value
    t = colorsys.hsv_to_rgb(h=h, s=s, v=v)
    return RGB(
        red=round(t[0] * MAX_COLOR),
        green=round(t[1] * MAX_COLOR),
        blue=round(t[2] * MAX_COLOR),
    )


def rgb_to_hsl(c: RGB) -> HSL:
    """
    Converts red, green, blue to hue, saturation, value

    Args:
        c (RGB): instance containing red, green, blue

    Returns:
        HSL: instance containing hue, saturation, lightness
    """
    r = float(c.red / MAX_COLOR)
    g = float(c.green / MAX_COLOR)
    b = float(c.blue / MAX_COLOR)
    t = colorsys.rgb_to_hls(r=r, g=g, b=b)
    return HSL(hue=t[0], saturation=t[2], lightness=t[1])


def hsv_to_hsl(c: HSV) -> HSL:
    """
    Convert hue, saturation, value to hue, saturation, lightness

    Args:
        c (HSV): instance containing hue, saturation, value

    Returns:
        HSL: instance containing hue, saturation, lightness
    """
    h = c.hue
    s = c.saturation
    v = c.value
    l = 0.5 * v * (2 - s)
    s = v * s / (1 - math.fabs(2 * l - 1))
    return HSL(h, s, l)


def hsl_to_hsv(c: HSL) -> HSV:
    """
    Convert hue, saturation, lightness to hue, saturation, value

    Args:
        c (HSL): instance containing hue, saturation, lightness

    Returns:
        HSV: instance containing hue, saturation, value
    """
    h = c.hue
    s = c.saturation
    l = c.lightness
    v = (2 * l + s * (1 - math.fabs(2 * l - 1))) / 2
    s = 2 * (v - l) / v
    return HSV(h, s, v)


def rgb_to_hex(rgb: RGB) -> str:
    """
    Converts rgb colors to int

    Args:
        rgb (color): Tuple of int with values from 0 to 255

    Returns:
        str: rgb as hex string
    """
    if len(rgb) != 3:
        raise ValueError("rgb must be a tuple of 3 integers")
    for i in rgb:
        if i < 0:
            raise ValueError("rgb contains a negative value")
        if i > MAX_COLOR:
            raise ValueError("rgb contains a value that is greater than MAX_COLOR")
    return "%02x%02x%02x" % rgb


def rgb_to_int(rgb: RGB) -> int:
    """
    Converts rgb colors to int

    Args:
        rgb (color): Tuple of int with values from 0 to 255

    Returns:
        int: rgb as int
    """
    return int(rgb_to_hex(rgb), 16)


def int_to_rgb(rgb_int: int) -> RGB:
    """
    Converts an integer that represents a rgb color into rgb object.

    Args:
        rgb_int (int): int that represents rgb color

    Returns:
        rgb: rgb with red, green and blue properties.
    """
    blue = rgb_int & MAX_COLOR
    green = (rgb_int >> 8) & MAX_COLOR
    red = (rgb_int >> 16) & MAX_COLOR
    return RGB(red, green, blue)


@overload
def lighten(rgb_color: int, percent: numbers.Number) -> RGB:
    ...


@overload
def lighten(rgb_color: RGB, percent: numbers.Number) -> RGB:
    ...


def lighten(rgb_color: Union[RGB, int], percent: numbers.Number) -> RGB:
    """
    Lightens an RGB instance

    Args:
        rgb_color (RGB | int): instance containing data
        percent (Number): Amount between 0 and 100 int lighten rgb by.

    Raises:
        ValueError: if percent is out of range

    Returns:
        RGB: RGB instance with lightened values applied.
    """
    if percent < 0 or percent > 100:
        raise ValueError("percent is expected to be between 0 and 100")
    # https://mdigi.tools/lighten-color
    # https://pastebin.com/KBAbAPh0
    if isinstance(rgb_color, int):
        _rgb = int_to_rgb(rgb_color)
    else:
        _rgb = rgb_color
    c_hsl = rgb_to_hsl(_rgb)
    amt = (percent * ((1 - c_hsl.lightness) * 100)) / 100
    l = c_hsl.lightness
    l += amt / 100
    increase = clamp(l, 0, 1)
    c2_hsl = HSL(c_hsl.hue, c_hsl.saturation, increase)
    c_rgb = hsl_to_rgb(c2_hsl)
    return c_rgb


@overload
def darken(rgb_color: int, percent: numbers.Number) -> RGB:
    ...


@overload
def darken(rgb_color: RGB, percent: numbers.Number) -> RGB:
    ...


def darken(rgb_color: Union[RGB, int], percent: numbers.Number) -> RGB:
    """
    Darkens an rgb instance

    Args:
        rgb_color (rgb | int): instance containing data
        percent (Number): Amount between 0 and 100 int darken rgb by.

    Raises:
        ValueError: if percent is out of range

    Returns:
        rgb: rgb instance with darkened values applied.
    """
    if percent < 0 or percent > 100:
        raise ValueError("percent is expected to be between 0 and 100")
    # https://mdigi.tools/lighten-color
    # https://pastebin.com/KBAbAPh0
    if isinstance(rgb_color, int):
        _rgb = int_to_rgb(rgb_color)
    else:
        _rgb = rgb_color
    c_hsl = rgb_to_hsl(_rgb)
    amt = (percent * (c_hsl.lightness * 100)) / 100
    l = c_hsl.lightness
    l -= amt / 100
    decrease = clamp(l, 0, 1)
    c2_hsl = HSL(c_hsl.hue, c_hsl.saturation, decrease)
    c_rgb = hsl_to_rgb(c2_hsl)
    return c_rgb



# coding: utf-8
# region Imports
from __future__ import annotations
from logging import exception
from typing import TYPE_CHECKING, Iterable, overload, Any
from enum import Enum, IntEnum
import uno

from com.sun.star.accessibility import XAccessible
from com.sun.star.accessibility import XAccessibleContext
from com.sun.star.awt import PosSize  # const
from com.sun.star.awt import WindowAttribute  # const
from com.sun.star.awt import VclWindowPeerAttribute  # const
from com.sun.star.awt import XExtendedToolkit
from com.sun.star.awt import XMenuBar
from com.sun.star.awt import XMessageBox
from com.sun.star.awt import XSystemDependentWindowPeer
from com.sun.star.awt import XToolkit
from com.sun.star.awt import XTopWindow
from com.sun.star.awt import XTopWindow2
from com.sun.star.awt import XUserInputInterception
from com.sun.star.awt import XWindow
from com.sun.star.awt import XWindow2
from com.sun.star.awt import XWindowPeer
from com.sun.star.beans import XPropertySet
from com.sun.star.container import XIndexContainer
from com.sun.star.frame import XDispatchProviderInterception
from com.sun.star.frame import XLayoutManager
from com.sun.star.frame import XFrame
from com.sun.star.frame import XFrame2
from com.sun.star.frame import XFramesSupplier
from com.sun.star.frame import XModel
from com.sun.star.lang import SystemDependent  # const
from com.sun.star.lang import XComponent
from com.sun.star.view import XControlAccess
from com.sun.star.view import XSelectionSupplier
from com.sun.star.ui import UIElementType  # const
from com.sun.star.ui import XImageManager
from com.sun.star.ui import XUIConfigurationManagerSupplier
from com.sun.star.ui import XUIConfigurationManager

if TYPE_CHECKING:
    from com.sun.star.frame import XController
    from com.sun.star.ui import XUIElement

from ..dialog import input as mInput
from ..utils import lo as mLo
from ..utils import props as mProps
from ..utils import info as mInfo
from ..utils import file_io as mFileIO
from ..utils import sys_info as m_sys_info
from ..exceptions import ex as mEx

from ooo.dyn.awt.rectangle import Rectangle
from ooo.dyn.awt.window_descriptor import WindowDescriptor
from ooo.dyn.awt.window_class import WindowClass

# endregion Imports

SysInfo = m_sys_info.SysInfo


class GUI:

    # region Class Enums
    # view settings zoom constants
    # ZoomEnum = DocumentZoomTypeEnum
    class ZoomEnum(IntEnum):
        OPTIMAL = 0
        """
        The page content width (excluding margins) at the current selection is fit into the view.
        """
        PAGE_WIDTH = 1
        """
        The page width at the current selection is fit into the view.
        """
        ENTIRE_PAGE = 2
        """
        A complete page of the document is fit into the view.
        """
        BY_VALUE = 3
        """
        The zoom is relative and is to be set via the property ViewSettings.ZoomValue.
        """
        PAGE_WIDTH_EXACT = 4
        """
        The page width at the current selection is fit into the view, with the view ends exactly at the end of the page.
        """
        ZOOM_50_PERCENT = 1000
        """Zoom 50%"""
        ZOOM_75_PERCENT = 1001
        """Zoom 75%"""
        ZOOM_100_PERCENT = 1002
        """Zoom 100%"""
        ZOOM_150_PERCENT = 1003
        """Zoom 150%"""
        ZOOM_200_PERCENT = 1004
        """Zoom 2000%"""

    # endregion Class Enums

    # region class Constants
    MENU_BAR = "private:resource/menubar/menubar"
    STATUS_BAR = "private:resource/statusbar/statusbar"
    FIND_BAR = "private:resource/toolbar/findbar"
    STANDARD_BAR = "private:resource/toolbar/standardbar"
    TOOL_BAR = "private:resource/toolbar/toolbar"

    class ToolBarName(str, Enum):
        THREE_D_OBJECTS_BAR = "3dobjectsbar"
        ADDON_LIBRE_LOGO_OFFICE_TOOL = "addon_LibreLogo.OfficeToolBar"
        ALIGNMENT_BAR = "alignmentbar"
        ARROWS_BAR = "arrowsbar"
        ARROW_SHAPES = "arrowshapes"
        BASIC_SHAPES = "basicshapes"
        BEZIER_OBJECT_BAR = "bezierobjectbar"
        CALL_OUT_SHAPES = "calloutshapes"
        CHANGES = "changes"
        CHOOSE_MODE_BAR = "choosemodebar"
        COLOR_BAR = "colorbar"
        COMMENTS_BAR = "commentsbar"
        COMMON_TASK_BAR = "commontaskbar"
        CONNECTORS_BAR = "connectorsbar"
        CUSTOM_TOOLBAR_1 = "custom_toolbar_1"
        DATASTREAMS = "datastreams"
        DESIGN_OBJECT_BAR = "designobjectbar"
        DIALOG_BAR = "dialogbar"
        DRAW_BAR = "drawbar"
        DRAWING_OBJECT_BAR = "drawingobjectbar"
        DRAW_OBJECT_BAR = "drawobjectbar"
        DRAW_TEXT_OBJECT_BAR = "drawtextobjectbar"
        ELLIPSES_BAR = "ellipsesbar"
        EXTRUSION_OBJECT_BAR = "extrusionobjectbar"
        FIND_BAR = "findbar"
        FLOW_CHART_SHAPES = "flowchartshapes"
        FONT_WORK_OBJECT_BAR = "fontworkobjectbar"
        FONT_WORK_SHAPE_TYPE = "fontworkshapetype"
        FORMAT_OBJECT_BAR = "formatobjectbar"
        FORMATTING = "Formatting"
        FORM_CONTROLS = "formcontrols"
        FORM_CONTROLS_BAR = "formcontrolsbar"
        FORM_DESIGN = "formdesign"
        FORM_OBJECT_BAR = "formobjectbar"
        FORMS_FILTER_BAR = "formsfilterbar"
        FORMS_NAVIGATION_BAR = "formsnavigationbar"
        FORM_TEXT_OBJECT_BAR = "formtextobjectbar"
        FRAME_OBJECT_BAR = "frameobjectbar"
        FULL_SCREEN_BAR = "fullscreenbar"
        GLUE_POINTS_OBJECT_BAR = "gluepointsobjectbar"
        GRAF_FILTER_BAR = "graffilterbar"
        GRAPH_ICO_OBJECT_BAR = "graphicobjectbar"
        INSERT_BAR = "insertbar"
        INSERT_CELLS_BAR = "insertcellsbar"
        INSERT_CONTROLS_BAR = "insertcontrolsbar"
        INSERT_OBJECT_BAR = "insertobjectbar"
        LINES_BAR = "linesbar"
        MACRO_BAR = "macrobar"
        MASTER_VIEW_TOOLBAR = "masterviewtoolbar"
        MEDIA_OBJECT_BAR = "mediaobjectbar"
        MORE_FORM_CONTROLS = "moreformcontrols"
        NAVIGATION_OBJECT_BAR = "navigationobjectbar"
        NUM_OBJECT_BAR = "numobjectbar"
        OLE_OBJECT_BAR = "oleobjectbar"
        OPTIMIZE_TABLE_BAR = "optimizetablebar"
        OPTIONS_BAR = "optionsbar"
        OUTLINE_TOOLBAR = "outlinetoolbar"
        POSITION_BAR = "positionbar"
        PREVIEW_BAR = "previewbar"
        PREVIEW_OBJECT_BAR = "previewobjectbar"
        QUERY_OBJECT_BAR = "queryobjectbar"
        RECTANGLES_BAR = "rectanglesbar"
        REPORT_CONTROLS = "reportcontrols"
        REPORT_OBJECT_BAR = "reportobjectbar"
        RESIZE_BAR = "resizebar"
        SECTION_ALIGNMENT_BAR = "sectionalignmentbar"
        SECTION_SHRINK_BAR = "sectionshrinkbar"
        SLIDE_VIEW_OBJECT_BAR = "slideviewobjectbar"
        SLIDE_VIEW_TOOL_BAR = "slideviewtoolbar"
        SQL_OBJECT_BAR = "sqlobjectbar"
        STANDARD_BAR = "standardbar"
        STAR_SHAPES = "starshapes"
        SYMBOL_SHAPES = "symbolshapes"
        TABLE_OBJECT_BAR = "tableobjectbar"
        TEXT_BAR = "textbar"
        TEXT_OBJECT_BAR = "textobjectbar"
        TOOLBAR = "toolbar"
        TRANSLATION_BAR = "translationbar"
        VIEWER_BAR = "viewerbar"
        ZOOM_BAR = "zoombar"

        def __str__(self) -> str:
            return self.value

    class SpecialWindows(str, Enum):
        BASIC_IDE = "BASICIDE"
        WELCOME_SCREEN = "WELCOMESCREEN"

        def __str__(self) -> str:
            return self.value

    class WindowSubtypes(str, Enum):
        BASE_TABLE = "BASETABLE"
        BASE_QUERY = "BASEQUERY"
        BASE_REPORT = "BASEREPORT"
        BASE_DIAGRAM = "BASEDIAGRAM"

        def __str__(self) -> str:
            return self.value

    class WindowInfo:
        __slots__ = (
            "component",
            "component",
            "frame",
            "window_name",
            "window_title",
            "window_file_name",
            "document_type",
        )

        def __init__(self) -> None:
            self.component: XComponent | None = None  # com.sun.star.lang.XComponent
            self.frame: XFrame | None = None  # com.sun.star.comp.framework.Frame
            self.window_name: str = ""  # Object Name
            self.window_title: str = ""  # Only mean to identify new documents
            self.window_file_name: str = ""  # URL of file name
            self.document_type: mLo.Lo.DocType = mLo.Lo.DocType.UNKNOWN  # Writer, Calc, ...

    # endregion class Constants

    # region ---------------- toolbar addition -------------------------

    @classmethod
    def get_toobar_resource(cls, name: GUI.ToolBarName | str) -> str:
        """
        Get toolbar resource for name

        Args:
            name (ToolBarName| str): Name of toolbar resource

        Returns:
            str: A formatted resource string such as ``private:resource/toolbar/zoombar``

        Note:
        """
        resource = f"private:resource/toolbar/{name}"
        return resource

    @classmethod
    def add_item_to_toolbar(cls, doc: XComponent, toolbar_name: str, item_name: str, im_fnm: str) -> None:
        """
        Add a user-defined icon and command to the start of the specified toolbar.

        Args:
            doc (XComponent): office document
            toolbar_name (str): toolbar name
            item_name (str): item name
            im_fnm (str): image file path

        """
        from com.sun.star.graphic import XGraphicProvider

        def load_graphic_file(im_fnm: str):
            # this method is also in Images module.
            # images module currently does not run as macro.
            # Pillow not needed for this method so make it local
            gprovider = mLo.Lo.create_instance_mcf(XGraphicProvider, "com.sun.star.graphic.GraphicProvider")
            if gprovider is None:
                return None

            file_props = mProps.Props.make_props(URL=mFileIO.FileIO.fnm_to_url(im_fnm))
            return gprovider.queryGraphic(file_props)

        try:
            cmd = mLo.Lo.make_uno_cmd(item_name)
            conf_man: XUIConfigurationManager = cls.get_ui_config_manager_doc(doc)
            image_man = mLo.Lo.qi(XImageManager, conf_man.getImageManager())
            if image_man is None:
                raise mEx.MissingInterfaceError(XImageManager)
            cmds = (cmd,)
            img = load_graphic_file(im_fnm)
            if img is None:
                mLo.Lo.print(f"Unable to load graphics file: '{im_fnm}'")
                return
            pics = (img,)
            image_man.insertImages(0, cmds, pics)

            # add item to toolbar
            settings = conf_man.getSettings(toolbar_name, True)
            con_settings = mLo.Lo.qi(XIndexContainer, settings)
            if con_settings is None:
                raise mEx.MissingInterfaceError(XIndexContainer)
            item_props = mProps.Props.make_bar_item(cmd, item_name)
            con_settings.insertByIndex(0, item_props)
            conf_man.replaceSettings(toolbar_name, con_settings)
        except Exception as e:
            mLo.Lo.print(e)

    # endregion ------------- toolbar addition -------------------------

    # region ---------------- floating frame, message box --------------

    @staticmethod
    def create_floating_frame(title: str, x: int, y: int, width: int, height: int) -> XFrame:
        """
        Create a floating XFrame at the given position and size

        Args:
            title (str): Floating frame title
            x (int): Frame x position
            y (int): Frame y position
            width (int): Frame width
            height (int): Frame Height

        Raises:
            MissingInterfaceError: If required interface can not be obtained.

        Returns:
            XFrame: Floating frame
        """
        xtoolkit = mLo.Lo.create_instance_mcf(XToolkit, "com.sun.star.awt.Toolkit")
        if xtoolkit is None:
            raise mEx.MissingInterfaceError(XToolkit)
        desc = WindowDescriptor(Type=WindowClass.TOP, WindowServiceName="modelessdialog", ParentIndex=-1)

        desc.Bounds = Rectangle(x, y, width, height)
        desc.WindowAttributes = (
            WindowAttribute.BORDER
            + WindowAttribute.MOVEABLE
            + WindowAttribute.CLOSEABLE
            + WindowAttribute.SIZEABLE
            + VclWindowPeerAttribute.CLIPCHILDREN
        )

        xwindow_peer = xtoolkit.createWindow(desc)
        window = mLo.Lo.qi(XWindow, xwindow_peer)
        if window is None:
            raise mEx.MissingInterfaceError(XWindow)

        xframe = mLo.Lo.create_instance_mcf(XFrame, "com.sun.star.frame.Frame")
        if xframe is None:
            raise mEx.MissingInterfaceError(XFrame)

        xframe.setName(title)
        xframe.initialize(window)

        xframes_sup = mLo.Lo.qi(XFramesSupplier, mLo.Lo.get_desktop())
        if xframes_sup is None:
            raise mEx.MissingInterfaceError(XFramesSupplier)

        xframes = xframes_sup.getFrames()
        if xframes is None:
            raise mEx.MissingInterfaceError(XFramesSupplier, "No desktop frames found")
        else:
            xframes.append(xframe)

        window.setVisible(True)
        return xframe

    @classmethod
    def show_message_box(cls, title: str, message: str) -> None:
        """
        Shows a message box

        Args:
            title (str): Messagebox Title
            message (str): Message to display

        Raises:
            MissingInterfaceError: If required interface is not present.
        """
        xtoolkit = mLo.Lo.create_instance_mcf(XToolkit, "com.sun.star.awt.Toolkit")
        xwindow = cls.get_window()
        if xtoolkit is None or xwindow is None:
            return None
        xpeer = mLo.Lo.qi(XWindowPeer, xwindow)
        if xpeer is None:
            raise mEx.MissingInterfaceError(XWindowPeer)
        desc = WindowDescriptor(
            Type=WindowClass.MODALTOP,
            WindowServiceName="infobox",
            ParentIndex=-1,
            Parent=xpeer,
            Bounds=Rectangle(0, 0, 300, 200),
            WindowAttributes=WindowAttribute.BORDER | WindowAttribute.MOVEABLE | WindowAttribute.CLOSEABLE,
        )

        desc_peer = xtoolkit.createWindow(desc)
        if desc_peer is None:
            msg_box = mLo.Lo.qi(XMessageBox, desc_peer)
            if msg_box is None:
                raise mEx.MissingInterfaceError(XMessageBox)
            msg_box.CaptionText = title
            msg_box.MessageText = message
            msg_box.execute()

    @staticmethod
    def get_password(title: str, input_msg: str) -> str:
        """
        Prompts for a password.

        If inside of a Office window then an Office Dialog is displayed prompting  for a password.
        If no Office window is available and attempt is made to create a ``tkinter`` dialog window for password input.

        ``tkinter`` does not ship with integrated python in LibreOffice. For this reason it may not be possible to display a ``tkinter`` dialog.
        It will depend on how your virtual environment is set up. In most cases this will work on Linux but not on windows.

        Args:
            title (str): Title of input box
            input_msg (str): Message to display

        Raises:
            Exception if unable to build a dialog password input.

        Returns:
            str: password as string or empty string if password is not given.

        See Also:
            :py:class:`~.input.Input`
        """
        try:
            result = mInput.Input.get_input(title=title, msg=input_msg, is_password=True)
            return result
        except Exception:
            # may not be in a LibreOffice window
            pass

        # try a tkinter dialog. Not available in macro mode.
        # this also means may not work on windows when virtual environment
        # is set to LibreOffice python.exe
        try:
            from ..dialog.tk_input import Window

            pass_inst = Window(title=title, input_msg=input_msg, is_password=True)
            return pass_inst.get_input()
        except ImportError:
            pass
        raise Exception("Unable to access a GUI to create a password dialog box")

    # endregion ------------- floating frame, message box --------------

    # region ---------------- controller and frame ---------------------

    @staticmethod
    def get_current_controller(odoc: object) -> XController:
        """
        Gets controller from document

        Args:
            odoc (object): office document

        Raises:
            MissingInterfaceError: If required interface is not present.

        Returns:
            XController: controller
        """
        doc = mLo.Lo.qi(XComponent, odoc)
        if doc is None:
            raise mEx.MissingInterfaceError(XComponent)
        model = mLo.Lo.qi(XModel, doc)
        if model is None:
            raise mEx.MissingInterfaceError(XComponent)
        return model.getCurrentController()

    @classmethod
    def get_frame(cls, doc: XComponent) -> XFrame:
        """
        Gets frame from doc

        Args:
            doc (XComponent): office document

        Returns:
            XFrame: document frame.
        """
        xcontroler = cls.get_current_controller(doc)
        return xcontroler.getFrame()

    @classmethod
    def get_control_access(cls, doc: XComponent) -> XControlAccess:
        """
        Get control access from office document

        Args:
            doc (XComponent): office document

        Raises:
            MissingInterfaceError: If doc does not implement XControlAccess interface.

        Returns:
            XControlAccess: control access
        """
        ca = mLo.Lo.qi(XControlAccess, cls.get_current_controller(doc))
        if ca is None:
            raise mEx.MissingInterfaceError(XControlAccess)
        return ca

    @staticmethod
    def get_uii(doc: XComponent) -> XUserInputInterception:
        """
        Gets user input interception

        Args:
            doc (XComponent): office document

        Raises:
            MissingInterfaceError: If doc does not implement XUserInputInterception interface.

        Returns:
            XUserInputInterception: user input interception
        """
        result = mLo.Lo.qi(XUserInputInterception, doc)
        if result is None:
            raise mEx.MissingInterfaceError(XUserInputInterception)
        return result

    @classmethod
    def get_selection_supplier(cls, odoc: object) -> XSelectionSupplier:
        """
        Gets selection supplier

        Args:
            odoc (object): office document

        Raises:
            MissingInterfaceError: if odoc does not implement XComponent interface.
            MissingInterfaceError: if XSelectionSupplier interface instance is not obtained.

        Returns:
            XSelectionSupplier: Selection supplier
        """
        doc = mLo.Lo.qi(XComponent, odoc)
        if doc is None:
            raise mEx.MissingInterfaceError(XComponent, "Not an office document")
        xcontroler = cls.get_current_controller(doc)
        result = mLo.Lo.qi(XSelectionSupplier, xcontroler)
        if result is None:
            raise mEx.MissingInterfaceError(XSelectionSupplier)
        return result

    @classmethod
    def get_dpi(cls, doc: XComponent) -> XDispatchProviderInterception:
        """
        Gets Dispatch provider interception

        Args:
            doc (XComponent): office document

        Raises:
            MissingInterfaceError: if XDispatchProviderInterception interface instance is not obtained.

        Returns:
            XDispatchProviderInterception: Dispatch provider interception
        """
        xframe = cls.get_frame(doc)
        result = mLo.Lo.qi(XDispatchProviderInterception, xframe)
        if result is None:
            raise mEx.MissingInterfaceError(XDispatchProviderInterception)
        return result

    # endregion ---------------- controller and frame ------------------

    # region ---------------- Office container window ------------------
    @staticmethod
    def get_window_idenity(obj: Any) -> GUI.WindowInfo:
        """
        Gets Identity Info for window of an object

        Args:
            obj (Any): object that implements XComponent

        Returns:
            GUI.Window: Window Info
        """
        # Credits to Scriptforge for the inspiration of this method.
        win = GUI.WindowInfo()
        component = mLo.Lo.qi(XComponent, obj)
        if component is None:
            return win
        win.component = component
        try:
            implementation = mInfo.Info.get_implementation_name(obj=obj)
        except Exception as e:
            mLo.Lo.print("Unable to get implementation name:")
            mLo.Lo.print(f"  {e}")
            return win
        try:
            identifier = mInfo.Info.get_identifier(obj)
        except Exception as e:
            mLo.Lo.print(f"Unable to get identity for implementation: {implementation}")
            mLo.Lo.print(f"  {e}")
            return win
        if implementation == "com.sun.star.comp.basic.BasicIDE":
            win.window_name = str(GUI.SpecialWindows.BASIC_IDE)
        elif implementation == "com.sun.star.comp.dba.ODatabaseDocument":  # No identifier
            model = mLo.Lo.qi(XModel, obj, True)
            win.window_file_name = str(mProps.Props.get_value(name="URL", props=model.getArgs()))
            if len(win.window_file_name) > 0:
                win.window_name = str(mFileIO.FileIO.url_to_path(win.window_file_name))
            win.document_type = mLo.Lo.DocType.BASE
        elif implementation in (
            "org.openoffice.comp.dbu.ODatasourceBrowser",
            "org.openoffice.comp.dbu.OTableDesign",
            "org.openoffice.comp.dbu.OQueryDesign",
            "org.openoffice.comp.dbu.ORelationDesign",
            "com.sun.star.comp.sfx2.BackingComp",
        ):
            win.frame = component.Frame
            win.window_name = str(GUI.SpecialWindows.WELCOME_SCREEN)
        else:
            if len(identifier) > 0:
                # Do not use URL : it contains the TemplateFile when new documents are created from a template
                win.window_file_name = component.Location
                if len(win.window_file_name) > 0:
                    win.window_name = str(mFileIO.FileIO.url_to_path(win.window_file_name))
                if hasattr(obj, "Title"):
                    win.window_title = obj.Title
                if identifier in (
                    "com.sun.star.sdb.FormDesign",
                    "com.sun.star.sdb.TextReportDesign",
                    "com.sun.star.text.TextDocument",
                ):
                    win.document_type = mLo.Lo.DocType.WRITER
                elif identifier == "com.sun.star.sheet.SpreadsheetDocument":
                    win.document_type = mLo.Lo.DocType.CALC
                elif identifier == "com.sun.star.presentation.PresentationDocument":
                    win.document_type = mLo.Lo.DocType.IMPRESS
                elif identifier == "com.sun.star.drawing.DrawingDocument":
                    win.document_type = mLo.Lo.DocType.DRAW
                elif identifier == "com.sun.star.formula.FormulaProperties":
                    win.document_type = mLo.Lo.DocType.MATH
        if win.frame is None:
            if hasattr(obj, "CurrentController"):
                win.frame = obj.CurrentController.Frame

        return win

    @classmethod
    def get_active_window(cls, obj: Any) -> str:
        """
        Gets Active window as string

        Args:
            obj (Any): doc like object

        Returns:
            str: Active window as string of found; Otherwise, an empty string.

        See Also:
            :py:meth:`~.gui.GUI.get_active_window`
        """
        win = cls.get_window_idenity(obj)
        if len(win.window_file_name) > 0:
            return str(mFileIO.FileIO.url_to_path(win.window_file_name))
        if len(win.window_name) > 0:
            return win.window_name
        if len(win.window_title) > 0:
            return win.window_title
        return ""

    @classmethod
    def activate(cls, window: str | XComponent) -> None:
        """
        Activates window

        Args:
            window (str | XComponent): Window name as str or doc as XComponent

        See Also:
            - :py:meth:`~.gui.GUI.get_active_window`
            - :py:meth:`~.gui.GUI.minimize`
            - :py:meth:`~.gui.GUI.maximize`
        """
        if not isinstance(window, str):
            swin = cls.get_active_window(window)
        else:
            swin = window
        desktop = mLo.Lo.XSCRIPTCONTEXT.getDesktop()
        enm = desktop.getComponents().createEnumeration()
        while enm.hasMoreElements():
            o_comp = enm.nextElement()
            win = cls.get_window_idenity(o_comp)
            if (
                (len(win.window_file_name) > 0 and win.window_file_name == mFileIO.FileIO.fnm_to_url(swin))
                or (len(win.window_name) > 0 and win.window_name == swin)
                or (len(win.window_title) > 0 and win.window_title == swin)
            ):
                if win.frame is None:
                    break
                container = win.frame.getContainerWindow()
                if container is None:
                    break
                xwin2 = mLo.Lo.qi(XWindow2, container)
                if xwin2 is None:
                    break
                top2 = mLo.Lo.qi(XTopWindow2, container)
                if top2 is None:
                    break
                if not xwin2.isVisible():
                    xwin2.setVisible(True)
                top2.IsMinimized = False
                xwin2.setFocus()
                top2.toFront()
                break

    @overload
    @classmethod
    def get_window(cls) -> XWindow:
        """
        Gets window

        Returns:
            XWindow: window instance
        """
        ...

    @overload
    @classmethod
    def get_window(cls, doc: XComponent) -> XWindow:
        """
        Gets window

        Args:
            doc (XComponent): Ofice document

        Returns:
            XWindow: window instance
        """
        ...

    @classmethod
    def get_window(cls, doc: XComponent = None) -> XWindow | None:
        """
        Gets window

        Args:
            doc (XComponent): Office document

        Returns:
            XWindow: window instance
        """
        if doc is None:
            desktop = mLo.Lo.get_desktop()
            if desktop is None:
                return None
            frame = desktop.getCurrentFrame()
            if frame is None:
                return None
            return frame.getContainerWindow()
        else:
            xcontroller = cls.get_current_controller(doc)
            return xcontroller.getFrame().getContainerWindow()

    @overload
    @classmethod
    def set_visible(cls, is_visible: bool) -> None:
        """
        Set window visibility.

        Args:
            is_visible (bool): If True window is set visible; Otherwise, window is set invisible.
        """
        ...

    @overload
    @classmethod
    def set_visible(cls, is_visible: bool, odoc: object) -> None:
        """
        Set window visibility.

        Args:
            is_visible (bool): If True window is set visible; Otherwise, window is set invisible.
            odoc (object): office document
        """
        ...

    @classmethod
    def set_visible(cls, is_visible: bool, odoc: object = None) -> None:
        """
        Set window visibility.

        Args:
            is_visible (bool): If True window is set visible; Otherwise, window is set invisible.
            odoc (object): office document
        """
        if odoc is None:
            xwindow = cls.get_window()
        else:
            doc = mLo.Lo.qi(XComponent, odoc)
            if doc is None:
                return
            xwindow = cls.get_frame(doc).getContainerWindow()

        if xwindow is not None:
            xwindow.setVisible(is_visible)
            xwindow.setFocus()

    @classmethod
    def set_size_window(cls, doc: XComponent, width: int, height: int) -> None:
        """
        Sets window size

        Args:
            doc (XComponent): office document
            width (int): Width of window
            height (int): Height of window
        """
        xwindow = cls.get_window(doc)
        rect = xwindow.getPosSize()
        xwindow.setPosSize(rect.X, rect.Y, width, height - 30, PosSize.POSSIZE)

    @classmethod
    def set_pos_size(cls, doc: XComponent, x: int, y: int, width: int, height: int) -> None:
        """
        Sets window position and size

        Args:
            doc (XComponent): office document
            x (int): Window X position
            y (int): Window Y Position
            width (int): Window Width
            height (int): Window Height
        """
        xwindow = cls.get_window(doc)
        xwindow.setPosSize(x, y, width, height, PosSize.POSSIZE)

    @classmethod
    def get_pos_size(cls, doc: XComponent) -> Rectangle:
        """
        Gets window position and Size

        Args:
            doc (XComponent): office document

        Returns:
            Rectangle: Rectangle representing position and size
        """
        xwindow = cls.get_window(doc)
        return xwindow.getPosSize()

    @staticmethod
    def get_top_window() -> XTopWindow:
        """
        Gets top window

        Raises:
            MissingInterfaceError: If XExtendedToolkit interface can not be obtained
            MissingInterfaceError: If XTopWindow interface can not be obtained

        Returns:
            XTopWindow: top window
        """
        tk = mLo.Lo.create_instance_mcf(XExtendedToolkit, "com.sun.star.awt.Toolkit")
        if tk is None:
            raise mEx.MissingInterfaceError(XExtendedToolkit)
        top_win = tk.getActiveTopWindow()
        if top_win is None:
            raise mEx.MissingInterfaceError(XTopWindow)
        return top_win

    @classmethod
    def get_title_bar(cls) -> str:
        """
        Gets title bar from top window

        Returns:
            str: title bar text if found; Otherwise, Empty string.
        """
        try:
            top_win = cls.get_top_window()
            acc = mLo.Lo.qi(XAccessible, top_win)
            if acc is None:
                raise mEx.MissingInterfaceError(XAccessible, "Top window not accessible")
            acc_content = acc.getAccessibleContext()
            if acc_content is not None:
                return acc_content.getAccessibleName()
        except Exception:
            pass
        # could not get title using top window. maybe the window was not currently on top.
        # get the title from XFrame2.
        try:
            frm2 = mLo.Lo.qi(XFrame2, mLo.Lo.get_frame())
            if frm2 is not None:
                return frm2.Title
        except Exception:
            pass
        return ""

    @staticmethod
    def get_screen_size() -> Rectangle:
        """
        Get the work area as Rectangle

        Raises:
            mEx.MissingInterfaceError: If XToolkit interface can not be obtained

        Returns:
            Rectangle: Work Area.

        Note:
            Original java method used java to get area.
            Original method seemed to return effective size. (i.e. without Windows' taskbar)

            This implementation calls ``Toolkit.getWorkArea()``.

        See also:
            `Toolkit <https://api.libreoffice.org/docs/idl/ref/servicecom_1_1sun_1_1star_1_1awt_1_1Toolkit.html>`_
        """
        tk = mLo.Lo.create_instance_mcf(XToolkit, "com.sun.star.awt.Toolkit")
        if tk is None:
            raise mEx.MissingInterfaceError(XToolkit)
        return tk.getWorkArea()

    @staticmethod
    def print_rect(r: Rectangle) -> None:
        """
        Prints a rectangle to the console

        Args:
            r (Rectangle): Rectangle to print
        """
        mLo.Lo.print(f"Rectangle: ({r.X}, {r.Y}), {r.Width} -- {r.Height}")

    @overload
    @classmethod
    def get_window_handle(cls) -> int | None:
        ...

    @overload
    @classmethod
    def get_window_handle(cls, doc: XComponent) -> int | None:
        ...

    @classmethod
    def get_window_handle(cls, doc: XComponent | None = None) -> int | None:
        """
        Gets handle to a window

        Args:
            doc (XComponent): document to get window handle for.

        Returns:
            int | None: handle as int on success; Otherwise, None.

        Note:
            This method was part of original java lib but was only set to work with windows.
            An attempt is made to support Linux and Mac; However, not tested at this point.

            Use this method at your own risk.
        """
        win = cls.get_window(doc)
        if win is None:
            return None
        try:
            win_peer = mLo.Lo.qi(XSystemDependentWindowPeer, win)
            pid = tuple([0 for _ in range(8)])  # tuple of zero's
            info = SysInfo.get_platform()
            if info == SysInfo.PlatformEnum.WINDOWS:
                system_type = SystemDependent.SYSTEM_WIN32
            elif info == SysInfo.PlatformEnum.MAC:
                system_type = SystemDependent.SYSTEM_MAC
            elif info == SysInfo.PlatformEnum.LINUX:
                system_type = SystemDependent.SYSTEM_XWINDOW
            else:
                mLo.Lo.print("Unable to support, don't know this system.")
                return None
            handel = int(win_peer.getWindowHandle(pid, system_type))
            return handel
        except Exception as e:
            mLo.Lo.print("Error getting windows handle")
            mLo.Lo.print(f"  {e}")
        return None

    @staticmethod
    def set_look_feel() -> None:
        """
        This method is not supported. Part of Original java lib.

        Raises:
            NotImplementedError: Not supported
        """
        raise NotImplementedError

    # endregion ------------- Office container window ------------------

    # region ---------------- min/max ----------------------------------
    @classmethod
    def maximize(cls, odoc: object) -> None:
        """
        Maximizes Office window

        Args:
            odoc (object): Office document

        See Also:
            - :py:meth:`~.gui.GUI.minimize`
            - :py:meth:`~.gui.GUI.activate`
            - :py:meth:`~.gui.GUI.get_active_window`
        """
        try:
            cls.activate(odoc)
            top_win = cls.get_top_window()
        except mEx.MissingInterfaceError as e:
            mLo.Lo.print("Unable to get top window")
            print(f"  {e}")
            return
        top2 = mLo.Lo.qi(XTopWindow2, top_win)
        if top2 is None:
            mLo.Lo.print("Unable to get top window (2)")
            return
        top2.IsMaximized = True

    @classmethod
    def minimize(cls, odoc: object) -> None:
        """
        Minimizes Office window

        Args:
            odoc (object): Office document

        See Also:
            - :py:meth:`~.gui.GUI.maximize`
            - :py:meth:`~.gui.GUI.activate`
            - :py:meth:`~.gui.GUI.get_active_window`
        """
        try:
            top_win = cls.get_top_window()
        except mEx.MissingInterfaceError as e:
            mLo.Lo.print("Unable to get top window")
            print(f"  {e}")
            return
        top2 = mLo.Lo.qi(XTopWindow2, top_win)
        if top2 is None:
            mLo.Lo.print("Unable to get top window (2)")
            return
        if top2.IsMinimized == False:
            cls.set_visible(is_visible=True, odoc=odoc)
            top2.IsMinimized = True

    # endregion ------------- min/max ----------------------------------

    # region ---------------- zooming ----------------------------------
    @overload
    @classmethod
    def zoom(cls, view: GUI.ZoomEnum) -> None:
        ...

    @overload
    @classmethod
    def zoom(cls, view: GUI.ZoomEnum, value: int) -> None:
        ...

    @overload
    @classmethod
    def zoom(cls, *, value: int = 0) -> None:
        ...

    @classmethod
    def zoom(cls, view: GUI.ZoomEnum = ZoomEnum.BY_VALUE, value: int = 0) -> None:
        """
        Sets document zoom level.

        Args:
            view (ZoomEnum): Zoom value
            value (int): The amount to zoom. :abbreviation:`eg:` 160 zooms 160%
                ``value`` has a min value of 1 and a max value of 3000. If value is out of range then 100% is used.
        """
        if view == GUI.ZoomEnum.OPTIMAL:
            mLo.Lo.dispatch_cmd("ZoomOptimal")
        elif view == GUI.ZoomEnum.PAGE_WIDTH or view == GUI.ZoomEnum.PAGE_WIDTH_EXACT:
            mLo.Lo.dispatch_cmd("ZoomPageWidth")
        elif view == GUI.ZoomEnum.ENTIRE_PAGE:
            mLo.Lo.dispatch_cmd("ZoomPage")
        elif view == GUI.ZoomEnum.ZOOM_50_PERCENT:
            mLo.Lo.dispatch_cmd("Zoom50Percent")
        elif view == GUI.ZoomEnum.ZOOM_75_PERCENT:
            mLo.Lo.dispatch_cmd("Zoom75Percent")
        elif view == GUI.ZoomEnum.ZOOM_100_PERCENT:
            mLo.Lo.dispatch_cmd("Zoom100Percent")
        elif view == GUI.ZoomEnum.ZOOM_150_PERCENT:
            mLo.Lo.dispatch_cmd("Zoom150Percent")
        elif view == GUI.ZoomEnum.ZOOM_200_PERCENT:
            mLo.Lo.dispatch_cmd("Zoom200Percent")
        elif view == GUI.ZoomEnum.BY_VALUE:
            if value <= 0 or value > 3000:
                value = 100
            p_dic = {"Zoom.Value": value, "Zoom.ValueSet": 28703, "Zoom.Type": 0}
            props = mProps.Props.make_props(**p_dic)
            mLo.Lo.dispatch_cmd(cmd="Zoom", props=props)
        else:
            mLo.Lo.print("Zoom not reconized. Aborting zoom.")
            return
        # https://wiki.openoffice.org/wiki/Documentation/DevGuide/Drawings/Zooming

        mLo.Lo.delay(500)

    @overload
    @classmethod
    def zoom_value(cls, value: int) -> None:
        """
        Sets document custom zoom.

        Args:
            value (int): The amount to zoom. :abbreviation:`eg:` 160 zooms 160%
        """
        ...

    @overload
    @classmethod
    def zoom_value(cls, value: int, view: GUI.ZoomEnum) -> None:
        """
        Sets document custom zoom.

        Args:
            value (int): The amount to zoom. :abbreviation:`eg:` 160 zooms 160%
                ``value`` has a min value of 1 and a max value of 3000. If value is out of range then 100% is used.
            view (ZoomEnum): Type of zoom. If ``view`` is not ``ZoomEnum.BY_VALUE`` then ``value`` is ignored. Defaults to ``ZoomEnum.BY_VALUE``.
        """
        ...

    @classmethod
    def zoom_value(cls, value: int, view: GUI.ZoomEnum = ZoomEnum.BY_VALUE) -> None:
        """
        Sets document custom zoom.

        Args:
            value (int): The amount to zoom. :abbreviation:`eg:` 160 zooms 160%
            view (ZoomEnum): Type of zoom. If ``view`` is not ``ZoomEnum.BY_VALUE`` then ``value`` is ignored. Defaults to ``ZoomEnum.BY_VALUE``.
        """
        cls.zoom(view=view, value=value)

    # endregion ------------- zooming ----------------------------------

    # region ---------------- UI config manager ------------------------

    @staticmethod
    def get_ui_config_manager(doc: XComponent) -> XUIConfigurationManager:
        """
        Gets ui config manager

        Args:
            doc (XComponent): office document

        Raises:
            MissingInterfaceError: If XModel interface can not be obtained
            MissingInterfaceError: If XUIConfigurationManagerSupplier interface can not be obtained

        Returns:
            XUIConfigurationManager: ui config manager
        """
        xmodel = mLo.Lo.qi(XModel, doc)
        if xmodel is None:
            raise mEx.MissingInterfaceError(XModel)
        xsupplier = mLo.Lo.qi(XUIConfigurationManagerSupplier, xmodel)
        if xsupplier is None:
            raise mEx.MissingInterfaceError(XUIConfigurationManagerSupplier)
        return xsupplier.getUIConfigurationManager()

    @staticmethod
    def get_ui_config_manager_doc(doc: XComponent) -> XUIConfigurationManager:
        """
        Gets ui config manager base upon doc type reported by :py:meth:`.Info.doc_type_service`.

        Args:
            doc (XComponent): office document

        Raises:
            MissingInterfaceError: If XModel interface can not be obtained
            MissingInterfaceError: If XUIConfigurationManagerSupplier interface can not be obtained
            Exception: If unable to get XUIConfigurationManager from XUIConfigurationManagerSupplier instance.

        Returns:
            XUIConfigurationManager: ui config manager
        """
        doc_type = mInfo.Info.doc_type_service(doc)

        xmodel = mLo.Lo.qi(XModel, doc)
        if xmodel is None:
            raise mEx.MissingInterfaceError(XModel)
        xsupplier = mLo.Lo.qi(XUIConfigurationManagerSupplier, xmodel)
        if xsupplier is None:
            raise mEx.MissingInterfaceError(XUIConfigurationManagerSupplier)
        try:
            return xsupplier.getUIConfigurationManager(str(doc_type))
        except Exception as e:
            raise Exception(f"Could not create a config manager using '{doc_type}'") from e

    # region print_ui_cmds()

    @overload
    @classmethod
    def print_ui_cmds(cls, ui_elem_name: str, config_man: XUIConfigurationManager) -> None:
        """
        Prints ui elements matching ``ui_elem_name`` to console.

        Args:
            ui_elem_name (str): Name of ui element
            config_man (XUIConfigurationManager): configuration manager
        """
        ...

    @overload
    @classmethod
    def print_ui_cmds(cls, ui_elem_name: str, doc: XComponent) -> None:
        """
        Prints ui elements matching ``ui_elem_name`` to console.

        Args:
            ui_elem_name (str): Name of ui element
            doc (XComponent): office document
        """
        ...

    @classmethod
    def print_ui_cmds(cls, *args, **kwargs) -> None:
        """
        Prints ui elements matching ``ui_elem_name`` to console.

        Args:
            ui_elem_name (str): Name of ui element
            config_man (XUIConfigurationManager): configuration manager
            doc (XComponent): office document
        """
        ordered_keys = (1, 2)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("ui_elem_name", "config_man", "doc")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("print_ui_cmds() got an unexpected keyword argument")
            ka[1] = kwargs.get("ui_elem_name", None)
            keys = ("config_man", "doc")
            for key in keys:
                if key in kwargs:
                    ka[2] = kwargs[key]
                    break
            return ka

        if count != 2:
            raise TypeError("print_ui_cmds() got an invalid number of arguments")

        kargs = get_kwargs()

        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg
        obj = mLo.Lo.qi(XUIConfigurationManager, kargs[2])
        if obj is None:
            cls._print_ui_cmds2(ui_elem_name=kargs[1], doc=kargs[2])
        else:
            cls._print_ui_cmds1(ui_elem_name=kargs[1], config_man=kargs[2])

    @staticmethod
    def _print_ui_cmds1(
        ui_elem_name: str,
        config_man: XUIConfigurationManager,
    ) -> None:
        """print every command used by the toolbar whose resource name is uiElemName"""
        # see Also: https://wiki.openoffice.org/wiki/Documentation/DevGuide/ProUNO/Properties
        try:
            settings = config_man.getSettings(ui_elem_name, True)
            num_settings = settings.getCount()
            print(f"No. of slements in '{ui_elem_name}' toolbar: {num_settings}")

            for i in range(num_settings):
                # line from java
                # PropertyValue[] settingProps =  Lo.qi(PropertyValue[].class, settings.getByIndex(i));
                setting_props = mLo.Lo.qi(XPropertySet, settings.getByIndex(i))
                val = mProps.Props.get_value(name="CommandURL", props=setting_props)
                print(f"{i}) {mProps.Props.prop_value_to_string(val)}")
            print()
        except exception as e:
            print(e)

    @classmethod
    def _print_ui_cmds2(cls, ui_elem_name: str, doc: XComponent) -> None:
        config_man = cls.get_ui_config_manager(doc)
        if config_man is None:
            print("Cannot create configuration manager")
            return
        cls.print_ui_cmds()

    # endregion print_ui_cmds()

    # endregion ------------- UI config manager ------------------------

    # region ---------------- layout manager ---------------------------

    # region    get_layout_manager()
    @overload
    @classmethod
    def get_layout_manager(cls) -> XLayoutManager:
        """
        Gets layout manager

        Args:
            doc (XComponent): office document

        Returns:
            XLayoutManager: Layout manager
        """
        ...

    @overload
    @classmethod
    def get_layout_manager(cls, doc: XComponent) -> XLayoutManager:
        """
        Gets layout manager

        Args:
            doc (XComponent): office document

        Raises:
            Exception: If unable to get layout manager

        Returns:
            XLayoutManager: Layout manager
        """
        ...

    @classmethod
    def get_layout_manager(cls, doc: XComponent = None) -> XLayoutManager:
        """
        Gets layout manager

        Args:
            doc (XComponent): office document

        Raises:
            Exception: If unable to get layout manager

        Returns:
            XLayoutManager: Layout manager
        """
        try:
            if doc is None:
                desktop = mLo.Lo.get_desktop()
                frame = desktop.getCurrentFrame()
            else:
                frame = cls.get_frame(doc)

            if frame is None:
                raise Exception("No current frame")

            lm = None
            prop_set = mLo.Lo.qi(XPropertySet, frame)
            lm = mLo.Lo.qi(XLayoutManager, prop_set.getPropertyValue("LayoutManager"))
            if lm is None:
                raise mEx.MissingInterfaceError(XLayoutManager)
            return lm
        except Exception as e:
            raise Exception("Could not access layout manager") from e

    # endregion    get_layout_manager()

    # region    show_memu_bar()
    @overload
    @classmethod
    def show_memu_bar(cls) -> None:
        ...

    @overload
    @classmethod
    def show_memu_bar(cls, doc: XComponent) -> None:
        ...

    @classmethod
    def show_memu_bar(cls, doc: XComponent = None) -> None:
        """
        Shows the main menu bar

        Args:
            doc (XComponent): doc (XComponent): office document
        """
        lm = cls.get_layout_manager(doc=doc)
        lm.showElement(GUI.MENU_BAR)

    # endregion show_memu_bar()

    # region    hide_memu_bar()
    @overload
    @classmethod
    def hide_memu_bar(cls) -> None:
        ...

    @overload
    @classmethod
    def hide_memu_bar(cls, doc: XComponent) -> None:
        ...

    @classmethod
    def hide_memu_bar(cls, doc: XComponent = None) -> None:
        """
        Hides the main menu bar

        Args:
            doc (XComponent): doc (XComponent): office document
        """
        lm = cls.get_layout_manager(doc=doc)
        lm.hideElement(GUI.MENU_BAR)

    @staticmethod
    def toggle_menu_bar() -> None:
        """
        Toggles the main menu visibility.

        If the menu is visible then it is hidden.
        If it is hidden then it will be made visible.

        Note:
            Toggle is done dispatching command ``Menubar``.
        """
        mLo.Lo.dispatch_cmd("Menubar")

    # endregion hide_memu_bar()

    # region    print_u_is()
    @overload
    @classmethod
    def print_u_is(cls) -> None:
        """print the resource names of every toolbar used by desktop"""
        ...

    @overload
    @classmethod
    def print_u_is(cls, doc: XComponent) -> None:
        """
        Print to console the resource names of every toolbar used by doc

        Args:
            doc (XComponent): office document
        """
        ...

    @overload
    @classmethod
    def print_u_is(cls, lm: XLayoutManager) -> None:
        """
        Print to console the resource names of every toolbar used by layout manager

        Args:
            lm (XLayoutManager): Layout manager
        """
        ...

    @classmethod
    def print_u_is(cls, *args, **kwargs) -> None:
        """
        Print to console the resource names of every toolbar used by doc

        Args:
            lm (XLayoutManager): Layout manager
            doc (XComponent): office document
        """
        ordered_keys = ("first",)
        kargs = {}
        if "doc" in kwargs:
            kargs["first"] = kwargs["doc"]
        elif "lm" in kwargs:
            kargs["first"] = kwargs["lm"]
        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg
        k_len = len(kargs)
        if k_len > 1:
            print("invalid number of arguments for print_u_is()")
            return
        if k_len == 0:
            lm = cls.get_layout_manager()
        else:
            obj = mLo.Lo.qi(XLayoutManager, kargs["first"])
            if obj is None:
                lm = cls.get_layout_manager(kargs["first"])
            else:
                lm = kargs["first"]
        if lm is None:
            print("No layout manager found")
            return
        ui_elems = lm.getElements()
        print(f"No. of UI Elemtnts: {len(ui_elems)}")
        for el in ui_elems:
            print(f"  {el.ResourceURL}; {cls.get_ui_element_type_str(el.Type)}")
        print()

    # endregion print_u_is()

    @staticmethod
    def get_ui_element_type_str(t: int) -> str:
        """
        Converts constant value to element type string

        `UIElementType <https://api.libreoffice.org/docs/idl/ref/namespacecom_1_1sun_1_1star_1_1ui_1_1UIElementType.html>`_
        determines the type of a user interface element which is controlled by a layout manager.

        Args:
            t (int): UIElementType constant Value from 0 to 8

        Raises:
            TypeError: If t is not a int
            ValueError: If t is not a valid UIElementType constant.

        Returns:
            str: element type string
        """
        if not isinstance(t, int):
            raise TypeError("'t' is not an int")
        if t == UIElementType.UNKNOWN:
            return "unknown"
        if t == UIElementType.MENUBAR:
            return "menubar"
        if t == UIElementType.POPUPMENU:
            return "popup menu"
        if t == UIElementType.TOOLBAR:
            return "toolbar"
        if t == UIElementType.STATUSBAR:
            return "status bar"
        if t == UIElementType.FLOATINGWINDOW:
            return "floating window"
        if t == UIElementType.PROGRESSBAR:
            return "progress bar"
        if t == UIElementType.TOOLPANEL:
            return "tool panel"
        if t == UIElementType.DOCKINGWINDOW:
            return "docking window"
        if t == UIElementType.COUNT:
            return "count"

        raise ValueError("'t' is is not a valid UIElementType value")

    @classmethod
    def printAllUICommands(cls, doc: XComponent) -> None:
        """
        Prints all ui commands to console

        Args:
            doc (XComponent): office document
        """
        conf_man = cls.get_ui_config_manager_doc(doc)
        if conf_man is None:
            print("No configuration manager found")
            return
        lm = cls.get_layout_manager(doc)
        if lm is None:
            print("No layout manager found")
            return
        ui_elmes = lm.getElements()
        print(f"No. of UI Elements: {len(ui_elmes)}")
        for el in ui_elmes:
            name = el.ResourceURL
            print(f"--- {name} ---")
            cls._print_ui_cmds1(ui_elem_name=name, config_man=conf_man)

    @classmethod
    def show_one(cls, doc: XComponent, show_elem: str) -> None:
        """
        Leave only the single specified toolbar visible

        Args:
            doc (XComponent): office document
            show_elem (str): name of element to show only.
        """
        show_elems = [show_elem]
        cls.show_only(doc=doc, show_elems=[show_elems])

    @classmethod
    def show_only(cls, doc: XComponent, show_elems: Iterable[str]) -> None:
        """
        Leave only the specified toolbars visible

        Raises:
            Exception: if unable to get layout manager from doc

        Args:
            doc (XComponent): office document
            show_elems (Iterable[str]): Elements to show
        """
        lm = cls.get_layout_manager(doc)
        ui_elmes = lm.getElements()
        cls.hide_except(lm=lm, ui_elms=ui_elmes, show_elems=show_elems)

        for el_name in show_elems:  # these elems are not in lm
            lm.createElement(el_name)  # so need to be created & shown
            lm.showElement(el_name)
            mLo.Lo.print(f"{el_name} made visible")

    @staticmethod
    def hide_except(lm: XLayoutManager, ui_elms: Iterable[XUIElement], show_elms: Iterable[str]) -> None:
        """
        Hide all of ``ui_elms``, except ones in ``show_elms``;
        delete any strings that match in ``show_elms``.

        Args:
            lm (XLayoutManager): Layout Manager
            ui_elms (Iterable[XUIElement]): Elements
            show_elms (Iterable[str]): elements to show
        """
        for ui_elm in ui_elms:
            el_name = ui_elm.ResourceURL
            to_hide = True
            # show_elms_lst = list(show_elms)
            for el in show_elms:
                if el == el_name:
                    show_elms.remove(el)  # this elem is in lm so remove from show_elems
                    to_hide = False
                    break
            if to_hide:
                lm.hideElement(el_name)
                mLo.Lo.print(f"{el_name} hidden")

    @classmethod
    def show_none(cls, doc: XComponent) -> None:
        """
        Make all the toolbars invisible

        Raises:
            Exception: if unable to get layout manager from doc

        Args:
            doc (XComponent): office document.
        """
        lm = cls.get_layout_manager(doc)
        if lm is None:
            mLo.Lo.print("No layout manager found")
            return
        ui_elms = lm.getElements()
        for ui_elm in ui_elms:
            elem_name = ui_elm.ResourceURL
            lm.hideElement(elem_name)
            mLo.Lo.print(f"{elem_name} hidden")

    # endregion ------------- layout manager ---------------------------

    # region ---------------- menu bar ---------------------------------

    @classmethod
    def get_menubar(cls, lm: XLayoutManager) -> XMenuBar:
        """
        Get menu bar

        Args:
            lm (XLayoutManager): layout manager

        Raises:
            TypeError: If lm is None
            MissingInterfaceError: If a required interface cannot be obtained.

        Returns:
            XMenuBar: menu bar
        """
        if lm is None:
            raise TypeError("'lm' is None. No layout manager available for menu discovery")

        omenu_bar = lm.getElement(cls.MENU_BAR)
        props = mLo.Lo.qi(XPropertySet, omenu_bar)
        if bar is None:
            raise mEx.MissingInterfaceError(XPropertySet)

        bar = mLo.Lo.qi(XMenuBar, props.getPropertyValue("XMenuBar"))
        # the XMenuBar reference is a property of the menubar UI
        if bar is None:
            raise mEx.MissingInterfaceError(XMenuBar)

    @classmethod
    def get_menu_max_id(cls, bar: XMenuBar) -> int:
        """
        Scan through the IDs used by all the items in
        this menubar, and return the largest ID encountered.

        Args:
            bar (XMenuBar): Menu bar

        Returns:
            int: Largest menu bar id if found; Otherwise, -1
        """
        if bar is None:
            return -1

        item_count = bar.getItemCount()
        max_id = -1
        for i in range(item_count):
            id = bar.getItemId(i)
            if id > max_id:
                max_id = id

        return max_id

    # endregion ------------- menu bar ---------------------------------



def check(exp: bool, err_msg: str, msg2: str = "out of range") -> None:
    """
    Preforms a validation check

    Args:
        exp (bool): Boolean expression such as ``self.value > 3``
        err_msg (str): Error Message
        msg2 (str, optional): Extended error message. Defaults to "out of range".

    Raises:
        AssertionError: If Validation fails
    """
    if not exp:
        # print(f"Type failure: {err_msg} {msg2}")
        # Normally, you'd use assert to throw an exception
        raise AssertionError(f"Type failure: {err_msg} {msg2}")


from .kind_base import KindBase
from . import kind_helper


class DrawingShapeKind(KindBase):
    """
    Values used with ``com.sun.star.drawing.*``

    See Also:
        - `Drawing API <https://api.libreoffice.org/docs/idl/ref/namespacecom_1_1sun_1_1star_1_1drawing.html>`_
        - `OpenOffice Shape Types <https://wiki.openoffice.org/wiki/Documentation/DevGuide/Drawings/Shape_Types>`_
    """

    APPLET_SHAPE = "AppletShape"
    CAPTION_SHAPE = "CaptionShape"
    CHART_LEGEND = "ChartLegend"
    CHART_TITLE = "ChartTitle"
    CLOSED_BEZIER_SHAPE = "ClosedBezierShape"
    CONNECTOR_SHAPE = "ConnectorShape"
    CONTROL_SHAPE = "ControlShape"
    CUSTOM_SHAPE = "CustomShape"
    ELLIPSE_SHAPE = "EllipseShape"
    GRAPHIC_OBJECT_SHAPE = "GraphicObjectShape"
    GROUP_SHAPE = "GroupShape"
    LINE_SHAPE = "LineShape"
    MEASURE_SHAPE = "MeasureShape"
    MEDIA_SHAPE = "MediaShape"
    OLE2_SHAPE = "OLE2Shape"
    OPEN_BEZIER_SHAPE = "OpenBezierShape"
    PAGE_SHAPE = "PageShape"
    PLUGIN_SHAPE = "PluginShape"
    POLY_LINE_SHAPE = "PolyLineShape"
    POLY_POLYGON_BEZIER_SHAPE = "PolyPolygonBezierShape"
    POLY_POLYGON_SHAPE = "PolyPolygonShape"
    RECTANGLE_SHAPE = "RectangleShape"
    TEXT_SHAPE = "TextShape"

    # could not find MediaShape in api.
    # https://api.libreoffice.org/docs/idl/ref/namespacecom_1_1sun_1_1star_1_1drawing.html
    # however it can be found in examples.
    # https://ask.libreoffice.org/t/how-to-add-video-to-impress-with-python/33050/2?u=vib

    def to_namespace(self) -> str:
        """Gets full name-space value of instance"""
        return f"com.sun.star.drawing.{self.value}"

    @staticmethod
    def from_str(s: str) -> "DrawingShapeKind":
        """
        Gets an ``DrawingShapeKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``DrawingShapeKind`` instance.

        Returns:
            DrawingShapeKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, DrawingShapeKind)



from __future__ import annotations
from typing import TYPE_CHECKING
from ..mock import mock_g

if mock_g.DOCS_BUILDING:
    from ..mock import unohelper
else:
    import unohelper

from com.sun.star.view import XSelectionChangeListener

if TYPE_CHECKING:
    from com.sun.star.lang import EventObject


class XSelectionChangeAdapter(unohelper.Base, XSelectionChangeListener):
    """
    Makes it possible to receive an event when the current selection changes.

    This class is meant a parent class.
    """

    def selectionChanged(self, event: EventObject) -> None:
        """
        is called when the selection changes.

        You can get the new selection via XSelectionSupplier from ``com.sun.star.lang.EventObject.Source``.
        """
        pass

    def disposing(self, event: EventObject) -> None:
        """
        gets called when the broadcaster is about to be disposed.

        All listeners and all other objects, which reference the broadcaster
        should release the reference to the source. No method should be invoked
        anymore on this object ( including ``XComponent.removeEventListener()`` ).

        This method is called for every listener registration of derived listener
        interfaced, not only for registrations at XComponent.
        """
        # from com.sun.star.lang.XEventListener
        pass



# coding: utf-8
from __future__ import annotations
from typing import Any
from .sheet_args import AbstractSheetArgs
from ..cancel_event_args import AbstractCancelEventArgs


class AbstractSheetCancelArgs(AbstractCancelEventArgs, AbstractSheetArgs):
    __slots__ = ()

    def __init__(self, source: Any, cancel=False) -> None:
        """
        Constructor

        Args:
            source (Any): Event Source
            cancel (bool, optional): Cancel value. Defaults to False.
        """
        super().__init__(source=source, cancel=cancel)


class SheetCancelArgs(AbstractSheetCancelArgs):
    """
    Sheet Cancel Event Args
    """

    __slots__ = ("source", "_event_name", "event_data", "name", "index", "doc", "sheet", "cancel", "_event_source")

    @staticmethod
    def from_args(args: SheetCancelArgs) -> SheetCancelArgs:
        """
        Gets a new instance from existing instance

        Args:
            args (SheetCancelArgs): Existing Instance

        Returns:
            SheetCancelArgs: args
        """
        eargs = SheetCancelArgs(source=args.source)
        eargs._event_name = args.event_name
        eargs._event_source = args.event_source
        eargs.doc = args.doc
        eargs.event_data = args.event_data
        eargs.index = args.index
        eargs.name = args.name
        eargs.sheet = args.sheet
        eargs.cancel = args.cancel
        return eargs
